# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import string
from numba import cuda
from numba.cuda.core import config
import numpy as np
import pytest
from pytest import param


pytestmark = pytest.mark.skipif(
    condition=config.ENABLE_CUDASIM,
    reason="no reason to run benchmarks in the simulator",
)


@pytest.mark.parametrize(
    "array_func",
    [
        param(
            lambda: cuda.device_array(128, dtype=np.float32),
            id="device_array",
        ),
        param(
            lambda: pytest.importorskip("torch").empty(
                (128,),
                dtype=pytest.importorskip("torch").float32,
                device="cuda:0",
            ),
            id="torch",
        ),
        param(
            lambda: pytest.importorskip("cupy").empty(128, dtype=np.float32),
            id="cupy",
        ),
    ],
)
def test_one_arg(benchmark, array_func):
    @cuda.jit("void(float32[:])")
    def one_arg(arr1):
        return

    def bench(func, arr):
        for _ in range(100):
            func(arr)

    benchmark(bench, one_arg[128, 128], array_func())


@pytest.mark.parametrize(
    "array_func",
    [
        param(
            lambda: [
                cuda.device_array(128, dtype=np.float32)
                for _ in range(len(string.ascii_lowercase))
            ],
            id="device_array",
        ),
        param(
            lambda: [
                pytest.importorskip("torch").empty(
                    (128,),
                    dtype=pytest.importorskip("torch").float32,
                    device="cuda:0",
                )
                for _ in range(len(string.ascii_lowercase))
            ],
            id="torch",
        ),
        param(
            lambda: [
                pytest.importorskip("cupy").empty(128, dtype=np.float32)
                for _ in range(len(string.ascii_lowercase))
            ],
            id="cupy",
        ),
    ],
)
def test_many_args(benchmark, array_func):
    many_arrs = array_func()

    @cuda.jit("void({})".format(", ".join(["float32[:]"] * len(many_arrs))))
    def many_args(
        a,
        b,
        c,
        d,
        e,
        f,
        g,
        h,
        i,
        j,
        k,
        l,
        m,
        n,
        o,
        p,
        q,
        r,
        s,
        t,
        u,
        v,
        w,
        x,
        y,
        z,
    ):
        return

    def bench(func, *arrs):
        for _ in range(100):
            func(*arrs)

    benchmark(bench, many_args[128, 128], *many_arrs)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.testing import unittest, CUDATestCase
from numba.cuda.cudadrv.driver import _have_nvjitlink
from llvmlite import ir

import numpy as np
import os
from numba import cuda
from numba.cuda import HAS_NUMBA
from numba.cuda.testing import skip_on_standalone_numba_cuda
from numba.cuda import types
from numba.cuda import config

if config.ENABLE_CUDASIM:
    raise unittest.SkipTest("Simulator does not support extending types")

import inspect
import math
import pickle
import unittest


import numba
from numba import njit
from numba.cuda import cgutils, jit
from numba.cuda.tests.support import TestCase, override_config
from numba.cuda.typing.templates import AttributeTemplate
from numba.cuda.cudadecl import registry as cuda_registry
from numba.cuda.cudaimpl import lower_attr as cuda_lower_attr

from numba.core import errors
from numba.cuda.errors import LoweringError

from numba.cuda.extending import (
    type_callable,
    lower_builtin,
    overload,
    overload_method,
    intrinsic,
    _Intrinsic,
    register_jitable,
    core_models,
    typeof_impl,
    register_model,
    make_attribute_wrapper,
)

TEST_BIN_DIR = os.getenv("NUMBA_CUDA_TEST_BIN_DIR")
if TEST_BIN_DIR:
    test_device_functions_a = os.path.join(
        TEST_BIN_DIR, "test_device_functions.a"
    )
    test_device_functions_cubin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cubin"
    )
    test_device_functions_cu = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cu"
    )
    test_device_functions_fatbin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.fatbin"
    )
    test_device_functions_fatbin_multi = os.path.join(
        TEST_BIN_DIR, "test_device_functions_multi.fatbin"
    )
    test_device_functions_o = os.path.join(
        TEST_BIN_DIR, "test_device_functions.o"
    )
    test_device_functions_ptx = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ptx"
    )
    test_device_functions_ltoir = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ltoir"
    )


class Interval:
    """
    A half-open interval on the real number line.
    """

    def __init__(self, lo, hi):
        self.lo = lo
        self.hi = hi

    def __repr__(self):
        return "Interval(%f, %f)" % (self.lo, self.hi)

    @property
    def width(self):
        return self.hi - self.lo


if HAS_NUMBA:
    from numba import njit
else:
    njit = None


@njit
def interval_width(interval):
    return interval.width


@njit
def sum_intervals(i, j):
    return Interval(i.lo + j.lo, i.hi + j.hi)


class IntervalType(types.Type):
    def __init__(self):
        super().__init__(name="Interval")


interval_type = IntervalType()


@typeof_impl.register(Interval)
def typeof_interval(val, c):
    return interval_type


@type_callable(Interval)
def type_interval(context):
    def typer(lo, hi):
        if isinstance(lo, types.Float) and isinstance(hi, types.Float):
            return interval_type

    return typer


@register_model(IntervalType)
class IntervalModel(core_models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("lo", types.float64),
            ("hi", types.float64),
        ]
        core_models.StructModel.__init__(self, dmm, fe_type, members)


make_attribute_wrapper(IntervalType, "lo", "lo")
make_attribute_wrapper(IntervalType, "hi", "hi")


@lower_builtin(Interval, types.Float, types.Float)
def impl_interval(context, builder, sig, args):
    typ = sig.return_type
    lo, hi = args
    interval = cgutils.create_struct_proxy(typ)(context, builder)
    interval.lo = lo
    interval.hi = hi
    return interval._getvalue()


@cuda_registry.register_attr
class Interval_attrs(AttributeTemplate):
    key = IntervalType

    def resolve_width(self, mod):
        return types.float64


@cuda_lower_attr(IntervalType, "width")
def cuda_Interval_width(context, builder, sig, arg):
    lo = builder.extract_value(arg, 0)
    hi = builder.extract_value(arg, 1)
    return builder.fsub(hi, lo)


# -----------------------------------------------------------------------
# Define a function's typing and implementation using the classical
# two-step API


def func1(x=None):
    raise NotImplementedError


def type_func1_(context):
    def typer(x=None):
        if x in (None, types.none):
            # 0-arg or 1-arg with None
            return types.int32
        elif isinstance(x, types.Float):
            # 1-arg with float
            return x

    return typer


type_func1 = type_callable(func1)(type_func1_)


@lower_builtin(func1)
@lower_builtin(func1, types.none)
def func1_nullary(context, builder, sig, args):
    return context.get_constant(sig.return_type, 42)


@lower_builtin(func1, types.Float)
def func1_unary(context, builder, sig, args):
    def func1_impl(x):
        return math.sqrt(2 * x)

    return context.compile_internal(builder, func1_impl, sig, args)


# -----------------------------------------------------------------------
# Overload an already defined built-in function, extending it for new types.


def call_func1_nullary(res):
    res[0] = func1()


def call_func1_unary(x, res):
    res[0] = func1(x)


class TestExtending(CUDATestCase):
    def test_attributes(self):
        @cuda.jit
        def f(r, x):
            iv = Interval(x[0], x[1])
            r[0] = iv.lo
            r[1] = iv.hi

        x = np.asarray((1.5, 2.5))
        r = np.zeros_like(x)

        f[1, 1](r, x)

        np.testing.assert_equal(r, x)

    def test_property(self):
        @cuda.jit
        def f(r, x):
            iv = Interval(x[0], x[1])
            r[0] = iv.width

        x = np.asarray((1.5, 2.5))
        r = np.zeros(1)

        f[1, 1](r, x)

        np.testing.assert_allclose(r[0], x[1] - x[0])

    @skip_on_standalone_numba_cuda
    def test_extension_type_as_arg(self):
        @cuda.jit
        def f(r, x):
            iv = Interval(x[0], x[1])
            r[0] = interval_width(iv)

        x = np.asarray((1.5, 2.5))
        r = np.zeros(1)

        f[1, 1](r, x)

        np.testing.assert_allclose(r[0], x[1] - x[0])

    @skip_on_standalone_numba_cuda
    def test_extension_type_as_retvalue(self):
        @cuda.jit
        def f(r, x):
            iv1 = Interval(x[0], x[1])
            iv2 = Interval(x[2], x[3])
            iv_sum = sum_intervals(iv1, iv2)
            r[0] = iv_sum.lo
            r[1] = iv_sum.hi

        x = np.asarray((1.5, 2.5, 3.0, 4.0))
        r = np.zeros(2)

        f[1, 1](r, x)

        expected = np.asarray((x[0] + x[2], x[1] + x[3]))
        np.testing.assert_allclose(r, expected)


class TestExtendingLinkage(CUDATestCase):
    @unittest.skipUnless(TEST_BIN_DIR, "Necessary binaries are not available")
    def test_extension_adds_linkable_code(self):
        files = (
            (test_device_functions_a, cuda.Archive),
            (test_device_functions_cubin, cuda.Cubin),
            (test_device_functions_cu, cuda.CUSource),
            (test_device_functions_fatbin, cuda.Fatbin),
            (test_device_functions_o, cuda.Object),
            (test_device_functions_ptx, cuda.PTXSource),
            (test_device_functions_ltoir, cuda.LTOIR),
        )

        lto = _have_nvjitlink()

        for path, ctor in files:
            if ctor == cuda.LTOIR and not lto:
                # Don't try to test with LTOIR if LTO is not enabled
                continue

            with open(path, "rb") as f:
                code_object = ctor(f.read())

            def external_add(x, y):
                return x + y

            @type_callable(external_add)
            def type_external_add(context):
                def typer(x, y):
                    if x == types.uint32 and y == types.uint32:
                        return types.uint32

                return typer

            @lower_builtin(external_add, types.uint32, types.uint32)
            def lower_external_add(context, builder, sig, args):
                context.active_code_library.add_linking_file(code_object)
                i32 = ir.IntType(32)
                fnty = ir.FunctionType(i32, [i32, i32])
                fn = cgutils.get_or_insert_function(
                    builder.module, fnty, "add_cabi"
                )
                return builder.call(fn, args)

            @cuda.jit(lto=lto)
            def use_external_add(r, x, y):
                r[0] = external_add(x[0], y[0])

            r = np.zeros(1, dtype=np.uint32)
            x = np.ones(1, dtype=np.uint32)
            y = np.ones(1, dtype=np.uint32) * 2

            use_external_add[1, 1](r, x, y)

            np.testing.assert_equal(r[0], 3)

            @cuda.jit(lto=lto)
            def use_external_add_device(x, y):
                return external_add(x, y)

            @cuda.jit(lto=lto)
            def use_external_add_kernel(r, x, y):
                r[0] = use_external_add_device(x[0], y[0])

            r = np.zeros(1, dtype=np.uint32)
            x = np.ones(1, dtype=np.uint32)
            y = np.ones(1, dtype=np.uint32) * 2

            use_external_add_kernel[1, 1](r, x, y)

            np.testing.assert_equal(r[0], 3)

    def test_linked_called_through_overload(self):
        cu_code = cuda.CUSource("""
            extern "C" __device__
            int bar(int *out, int a)
            {
              *out = a * 2;
              return 0;
            }
        """)

        bar = cuda.declare_device("bar", "int32(int32)", link=cu_code)

        def bar_call(val):
            pass

        @overload(bar_call, target="cuda")
        def ol_bar_call(a):
            return lambda a: bar(a)

        @cuda.jit("void(int32[::1], int32[::1])")
        def foo(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = bar_call(x[i])

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        foo[1, 32](r, x)

        np.testing.assert_equal(r, x * 2)


class TestLowLevelExtending(TestCase):
    """
    Test the low-level two-tier extension API.
    """

    # Check with `@jit` from within the test process and also in a new test
    # process so as to check the registration mechanism.

    def test_func1(self):
        pyfunc = call_func1_nullary
        cfunc = jit(pyfunc)
        res = np.zeros(1)
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            cfunc[1, 1](res)
        self.assertPreciseEqual(res[0], 42.0)
        pyfunc = call_func1_unary
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            cfunc = jit(pyfunc)
        self.assertPreciseEqual(res[0], 42.0)
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            cfunc[1, 1](18.0, res)
        self.assertPreciseEqual(res[0], 6.0)

    @TestCase.run_test_in_subprocess
    def test_func1_isolated(self):
        self.test_func1()

    def test_type_callable_keeps_function(self):
        self.assertIs(type_func1, type_func1_)
        self.assertIsNotNone(type_func1)


class TestHighLevelExtending(TestCase):
    """
    Test the high-level combined API.
    """

    def test_typing_vs_impl_signature_mismatch_handling(self):
        """
        Tests that an overload which has a differing typing and implementing
        signature raises an exception.
        """

        def gen_ol(impl=None):
            def myoverload(a, b, c, kw=None):
                pass

            @overload(myoverload)
            def _myoverload_impl(a, b, c, kw=None):
                return impl

            @jit
            def foo(a, b, c, d):
                myoverload(a, b, c, kw=d)

            return foo

        sentinel = "Typing and implementation arguments differ in"

        # kwarg value is different
        def impl1(a, b, c, kw=12):
            if a > 10:
                return 1
            else:
                return -1

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl1)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("keyword argument default values", msg)
        self.assertIn('<Parameter "kw=12">', msg)
        self.assertIn('<Parameter "kw=None">', msg)

        # kwarg name is different
        def impl2(a, b, c, kwarg=None):
            if a > 10:
                return 1
            else:
                return -1

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl2)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("keyword argument names", msg)
        self.assertIn('<Parameter "kwarg=None">', msg)
        self.assertIn('<Parameter "kw=None">', msg)

        # arg name is different
        def impl3(z, b, c, kw=None):
            if a > 10:  # noqa: F821
                return 1
            else:
                return -1

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl3)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("argument names", msg)
        self.assertFalse("keyword" in msg)
        self.assertIn('<Parameter "a">', msg)
        self.assertIn('<Parameter "z">', msg)

        from .overload_usecases import impl4, impl5

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl4)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("argument names", msg)
        self.assertFalse("keyword" in msg)
        self.assertIn("First difference: 'z'", msg)

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl5)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("argument names", msg)
        self.assertFalse("keyword" in msg)
        self.assertIn('<Parameter "a">', msg)
        self.assertIn('<Parameter "z">', msg)

        # too many args
        def impl6(a, b, c, d, e, kw=None):
            if a > 10:
                return 1
            else:
                return -1

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl6)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("argument names", msg)
        self.assertFalse("keyword" in msg)
        self.assertIn('<Parameter "d">', msg)
        self.assertIn('<Parameter "e">', msg)

        # too few args
        def impl7(a, b, kw=None):
            if a > 10:
                return 1
            else:
                return -1

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl7)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("argument names", msg)
        self.assertFalse("keyword" in msg)
        self.assertIn('<Parameter "c">', msg)

        # too many kwargs
        def impl8(a, b, c, kw=None, extra_kwarg=None):
            if a > 10:
                return 1
            else:
                return -1

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl8)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("keyword argument names", msg)
        self.assertIn('<Parameter "extra_kwarg=None">', msg)

        # too few kwargs
        def impl9(a, b, c):
            if a > 10:
                return 1
            else:
                return -1

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(impl9)[1, 1](1, 2, 3, 4)
        msg = str(e.exception)
        self.assertIn(sentinel, msg)
        self.assertIn("keyword argument names", msg)
        self.assertIn('<Parameter "kw=None">', msg)

    def test_typing_vs_impl_signature_mismatch_handling_var_positional(self):
        """
        Tests that an overload which has a differing typing and implementing
        signature raises an exception and uses VAR_POSITIONAL (*args) in typing
        """

        def myoverload(a, kw=None):
            pass

        from .overload_usecases import var_positional_impl

        overload(myoverload)(var_positional_impl)

        @jit
        def foo(a, b):
            myoverload(a, b, 9, kw=11)

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                foo[1, 1](1, 5)
        msg = str(e.exception)
        self.assertIn("VAR_POSITIONAL (e.g. *args) argument kind", msg)
        self.assertIn("offending argument name is '*star_args_token'", msg)

    def test_typing_vs_impl_signature_mismatch_handling_var_keyword(self):
        """
        Tests that an overload which uses **kwargs (VAR_KEYWORD)
        """

        def gen_ol(impl, strict=True):
            def myoverload(a, kw=None):
                pass

            overload(myoverload, strict=strict)(impl)

            @jit
            def foo(a, b):
                myoverload(a, kw=11)

            return foo

        # **kwargs in typing
        def ol1(a, **kws):
            def impl(a, kw=10):
                return a

            return impl

        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            gen_ol(ol1, False)[1, 1](
                1, 2
            )  # no error if strictness not enforced
        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(ol1)[1, 1](1, 2)
        msg = str(e.exception)
        self.assertIn("use of VAR_KEYWORD (e.g. **kwargs) is unsupported", msg)
        self.assertIn("offending argument name is '**kws'", msg)

        # **kwargs in implementation
        def ol2(a, kw=0):
            def impl(a, **kws):
                return a

            return impl

        with self.assertRaises(errors.TypingError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                gen_ol(ol2)[1, 1](1, 2)
        msg = str(e.exception)
        self.assertIn("use of VAR_KEYWORD (e.g. **kwargs) is unsupported", msg)
        self.assertIn("offending argument name is '**kws'", msg)

    def test_overload_method_kwargs(self):
        # Issue #3489
        @overload_method(types.Array, "foo")
        def fooimpl(arr, a_kwarg=10):
            def impl(arr, a_kwarg=10):
                return a_kwarg

            return impl

        @jit
        def bar(A, res):
            res[0] = A.foo()
            res[1] = A.foo(20)
            res[2] = A.foo(a_kwarg=30)

        Z = np.arange(5)
        res = np.zeros(3)
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            bar[1, 1](Z, res)
        self.assertEqual(res[0], 10)
        self.assertEqual(res[1], 20)
        self.assertEqual(res[2], 30)

    def test_overload_method_literal_unpack(self):
        # Issue #3683
        @overload_method(types.Array, "litfoo")
        def litfoo(arr, val):
            # Must be an integer
            if isinstance(val, types.Integer):
                # Must not be literal
                if not isinstance(val, types.Literal):

                    def impl(arr, val):
                        return val

                    return impl

        @jit
        def bar(A, res):
            res[0] = A.litfoo(0xCAFE)

        A = np.zeros(1)
        res = np.zeros(1)
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            bar[1, 1](A, res)
        self.assertEqual(res[0], 0xCAFE)


def _assert_cache_stats(cfunc, expect_hit, expect_misses):
    hit = cfunc._cache_hits[cfunc.signatures[0]]
    if hit != expect_hit:
        raise AssertionError("cache not used")
    miss = cfunc._cache_misses[cfunc.signatures[0]]
    if miss != expect_misses:
        raise AssertionError("cache not used")


class TestIntrinsic(TestCase):
    def test_void_return(self):
        """
        Verify that returning a None from codegen function is handled
        automatically for void functions, otherwise raise exception.
        """

        @intrinsic
        def void_func(typingctx, a):
            sig = types.void(types.int32)

            def codegen(context, builder, signature, args):
                pass  # do nothing, return None, should be turned into
                # dummy value

            return sig, codegen

        @intrinsic
        def non_void_func(typingctx, a):
            sig = types.int32(types.int32)

            def codegen(context, builder, signature, args):
                pass  # oops, should be returning a value here, raise exception

            return sig, codegen

        @jit
        def call_void_func():
            void_func(1)

        @jit
        def call_non_void_func():
            non_void_func(1)

        # void func should work
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            self.assertEqual(call_void_func[1, 1](), None)
        # not void function should raise exception
        with self.assertRaises(LoweringError) as e:
            with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
                call_non_void_func[1, 1]()
        self.assertIn("non-void function returns None", e.exception.msg)

    def test_serialization(self):
        """
        Test serialization of intrinsic objects
        """

        # define a intrinsic
        @intrinsic
        def identity(context, x):
            def codegen(context, builder, signature, args):
                return args[0]

            sig = x(x)
            return sig, codegen

        # use in a jit function
        @jit
        def foo(x):
            identity(x)

        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            self.assertEqual(foo[1, 1](1), None)

        # get serialization memo
        memo = _Intrinsic._memo
        memo_size = len(memo)

        # pickle foo and check memo size
        serialized_foo = pickle.dumps(foo)
        # increases the memo size
        memo_size += 1
        self.assertEqual(memo_size, len(memo))
        # unpickle
        foo_rebuilt = pickle.loads(serialized_foo)
        self.assertEqual(memo_size, len(memo))
        # check rebuilt foo

        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            self.assertEqual(foo[1, 1](1), foo_rebuilt[1, 1](1))

        # pickle identity directly
        serialized_identity = pickle.dumps(identity)
        # memo size unchanged
        self.assertEqual(memo_size, len(memo))
        # unpickle
        identity_rebuilt = pickle.loads(serialized_identity)
        # must be the same object
        self.assertIs(identity, identity_rebuilt)
        # memo size unchanged
        self.assertEqual(memo_size, len(memo))

    def test_deserialization(self):
        """
        Test deserialization of intrinsic
        """

        def defn(context, x):
            def codegen(context, builder, signature, args):
                return args[0]

            return x(x), codegen

        memo = _Intrinsic._memo
        memo_size = len(memo)
        # invoke _Intrinsic indirectly to avoid registration which keeps an
        # internal reference inside the compiler
        original = _Intrinsic("foo", defn)
        self.assertIs(original._defn, defn)
        pickled = pickle.dumps(original)
        # by pickling, a new memo entry is created
        memo_size += 1
        self.assertEqual(memo_size, len(memo))
        del original  # remove original before unpickling

        # by deleting, the memo entry is NOT removed due to recent
        # function queue
        self.assertEqual(memo_size, len(memo))

        # Manually force clear of _recent queue
        _Intrinsic._recent.clear()
        memo_size -= 1
        self.assertEqual(memo_size, len(memo))

        rebuilt = pickle.loads(pickled)
        # verify that the rebuilt object is different
        self.assertIsNot(rebuilt._defn, defn)

        # the second rebuilt object is the same as the first
        second = pickle.loads(pickled)
        self.assertIs(rebuilt._defn, second._defn)

    def test_docstring(self):
        @intrinsic
        def void_func(typingctx, a: int):
            """void_func docstring"""
            sig = types.void(types.int32)

            def codegen(context, builder, signature, args):
                pass  # do nothing, return None, should be turned into
                # dummy value

            return sig, codegen

        self.assertEqual(
            "numba.cuda.tests.cudapy.test_extending", void_func.__module__
        )
        self.assertEqual("void_func", void_func.__name__)
        self.assertEqual(
            "TestIntrinsic.test_docstring.<locals>.void_func",
            void_func.__qualname__,
        )
        self.assertDictEqual({"a": int}, void_func.__annotations__)
        self.assertEqual("void_func docstring", void_func.__doc__)


class TestRegisterJitable(unittest.TestCase):
    def test_no_flags(self):
        @register_jitable
        def foo(x, y):
            x[0] += y

        def bar(x, y):
            foo(x, y)
            x[0] += x[0]

        cbar = jit(bar)

        x = np.array([1, 2])
        bar(x, 2)
        self.assertEqual(x[0], 6)
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            cbar[1, 1](x, 2)
        self.assertEqual(x[0], 16)


class TestOverloadPreferLiteral(TestCase):
    def test_overload(self):
        def prefer_lit(x):
            pass

        def non_lit(x):
            pass

        def ov(x):
            if isinstance(x, types.IntegerLiteral):
                # With prefer_literal=False, this branch will not be reached.
                if x.literal_value == 1:

                    def impl(x):
                        return 0xCAFE

                    return impl
                else:
                    raise errors.TypingError("literal value")
            else:

                def impl(x):
                    return x * 100

                return impl

        overload(prefer_lit, prefer_literal=True)(ov)
        overload(non_lit)(ov)

        @jit
        def check_prefer_lit(x, res):
            res[0] = prefer_lit(1)
            res[1] = prefer_lit(2)
            res[2] = prefer_lit(x)

        res = np.zeros(3)
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            check_prefer_lit[1, 1](3, res)
        a, b, c = res
        self.assertEqual(a, 0xCAFE)
        self.assertEqual(b, 200)
        self.assertEqual(c, 300)

        @jit
        def check_non_lit(x, res):
            res[0] = non_lit(1)
            res[1] = non_lit(2)
            res[2] = non_lit(x)

        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            check_non_lit[1, 1](3, res)
        a, b, c = res
        self.assertEqual(a, 100)
        self.assertEqual(b, 200)
        self.assertEqual(c, 300)


class TestNumbaInternalOverloads(TestCase):
    def test_signatures_match_overloaded_api(self):
        # This is a "best-effort" test to try and ensure that Numba's internal
        # overload declarations have signatures with argument names that match
        # the API they are overloading. The purpose of ensuring there is a
        # match is so that users can use call-by-name for positional arguments.

        # Set this to:
        # 0 to make violations raise a ValueError (default).
        # 1 to get violations reported to STDOUT
        # 2 to get a verbose output of everything that was checked and its state
        #   reported to STDOUT.
        DEBUG = 0

        # np.random.* does not have a signature exposed to `inspect`... so
        # custom parse the docstrings.
        def sig_from_np_random(x):
            if not x.startswith("_"):
                thing = getattr(np.random, x)
                if inspect.isbuiltin(thing):
                    docstr = thing.__doc__.splitlines()
                    for l in docstr:
                        if l:
                            sl = l.strip()
                            if sl.startswith(x):  # its the signature
                                # special case np.random.seed, it has `self` in
                                # the signature whereas all the other functions
                                # do not!?
                                if x == "seed":
                                    sl = "seed(seed)"

                                fake_impl = f"def {sl}:\n\tpass"
                                l = {}
                                try:
                                    exec(fake_impl, {}, l)
                                except SyntaxError:
                                    # likely elipsis, e.g. rand(d0, d1, ..., dn)
                                    if DEBUG == 2:
                                        print(
                                            "... skipped as cannot parse "
                                            "signature"
                                        )
                                    return None
                                else:
                                    fn = l.get(x)
                                    return inspect.signature(fn)

        def checker(func, overload_func):
            if DEBUG == 2:
                print(f"Checking: {func}")

            def create_message(func, overload_func, func_sig, ol_sig):
                msg = []
                s = (
                    f"{func} from module '{getattr(func, '__module__')}' "
                    "has mismatched sig."
                )
                msg.append(s)
                msg.append(f"    - expected: {func_sig}")
                msg.append(f"    -      got: {ol_sig}")
                lineno = inspect.getsourcelines(overload_func)[1]
                tmpsrcfile = inspect.getfile(overload_func)
                srcfile = tmpsrcfile.replace(numba.__path__[0], "")
                msg.append(f"from {srcfile}:{lineno}")
                msgstr = "\n" + "\n".join(msg)
                return msgstr

            func_sig = None
            try:
                func_sig = inspect.signature(func)
            except ValueError:
                # probably a built-in/C code, see if it's a np.random function
                if fname := getattr(func, "__name__", False):
                    if maybe_func := getattr(np.random, fname, False):
                        if maybe_func == func:
                            # it's a built-in from np.random
                            func_sig = sig_from_np_random(fname)

            if func_sig is not None:
                ol_sig = inspect.signature(overload_func)
                x = list(func_sig.parameters.keys())
                y = list(ol_sig.parameters.keys())
                for a, b in zip(x[: len(y)], y):
                    if a != b:
                        p = func_sig.parameters[a]
                        if p.kind == p.POSITIONAL_ONLY:
                            # probably a built-in/C code
                            if DEBUG == 2:
                                print(
                                    "... skipped as positional only "
                                    "arguments found"
                                )
                            break
                        elif "*" in str(p):  # probably *args or similar
                            if DEBUG == 2:
                                print("... skipped as contains *args")
                            break
                        else:
                            # Only error/report on functions that have a module
                            # or are from somewhere other than Numba.
                            if (
                                not func.__module__
                                or not func.__module__.startswith("numba")
                            ):
                                msgstr = create_message(
                                    func, overload_func, func_sig, ol_sig
                                )
                                if DEBUG != 0:
                                    if DEBUG == 2:
                                        print("... INVALID")
                                    if msgstr:
                                        print(msgstr)
                                    break
                                else:
                                    raise ValueError(msgstr)
                            else:
                                if DEBUG == 2:
                                    if not func.__module__:
                                        print(
                                            "... skipped as no __module__ "
                                            "present"
                                        )
                                    else:
                                        print("... skipped as Numba internal")
                                break
                else:
                    if DEBUG == 2:
                        print("... OK")

        # Compile something to make sure that the typing context registries
        # are populated with everything from the CPU target.
        jit(lambda: None).compile(())
        tyctx = numba.cuda.target.CUDATypingContext()
        tyctx.refresh()

        # Walk the registries and check each function that is an overload
        regs = tyctx._registries
        for k, v in regs.items():
            for item in k.functions:
                if getattr(item, "_overload_func", False):
                    checker(item.key, item._overload_func)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
import numpy as np

from numba.cuda.tests.support import TestCase, MemoryLeakMixin
from numba import cuda
from numba.cuda.testing import skip_on_cudasim
from numba.cuda.misc.special import literal_unroll
from numba.cuda import config


@skip_on_cudasim("doesn't work in the simulator")
class TestArrayReductions(MemoryLeakMixin, TestCase):
    """
    Test array reduction methods and functions such as .sum(), .max(), etc.
    """

    def setUp(self):
        super(TestArrayReductions, self).setUp()
        np.random.seed(42)
        self.old_nrt_setting = config.CUDA_ENABLE_NRT
        self.old_perf_warnings_setting = config.DISABLE_PERFORMANCE_WARNINGS
        config.CUDA_ENABLE_NRT = True
        config.DISABLE_PERFORMANCE_WARNINGS = 1

    def tearDown(self):
        config.CUDA_ENABLE_NRT = self.old_nrt_setting
        config.DISABLE_PERFORMANCE_WARNINGS = self.old_perf_warnings_setting
        super(TestArrayReductions, self).tearDown()

    def test_all_basic(self):
        cases = (
            np.float64([1.0, 0.0, float("inf"), float("nan")]),
            np.float64([1.0, -0.0, float("inf"), float("nan")]),
            np.float64([1.0, 1.5, float("inf"), float("nan")]),
            np.float64([[1.0, 1.5], [float("inf"), float("nan")]]),
            np.float64([[1.0, 1.5], [1.5, 1.0]]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for case in literal_unroll(cases):
                out[i] = np.all(case)
                i += 1

        expected = np.array([np.all(a) for a in cases], dtype=np.bool_)
        out = cuda.to_device(np.zeros(len(cases), dtype=np.bool_))
        kernel[1, 1](out)
        got = out.copy_to_host()
        self.assertPreciseEqual(expected, got)

    def test_any_basic(self):
        cases = (
            np.float64([0.0, -0.0, 0.0, 0.0]),
            np.float64([0.0, -0.0, np.nan, 0.0]),
            np.float64([0.0, -0.0, float("inf"), 0.0]),
            np.float64([0.0, -0.0, 1.5, 0.0]),
            np.float64([[0.0, -0.0], [1.5, 0.0]]),
            np.float64([[0.0, -0.0], [1.5, 0.0]])[::-1],
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(cases):
                out[i] = np.any(arr)
                i += 1

        expected = np.array([np.any(a) for a in cases], dtype=np.bool_)
        out = cuda.to_device(np.zeros(len(cases), dtype=np.bool_))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_sum_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64(
                [np.nan, -1.5, 2.5, np.nan, float("inf"), -float("inf"), 3.0]
            ),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.sum(arr)
                i += 1

        expected = np.array([np.sum(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_mean_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64(
                [np.nan, -1.5, 2.5, np.nan, float("inf"), -float("inf"), 3.0]
            ),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.mean(arr)
                i += 1

        expected = np.array([np.mean(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_var_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64(
                [np.nan, -1.5, 2.5, np.nan, float("inf"), -float("inf"), 3.0]
            ),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.var(arr)
                i += 1

        expected = np.array([np.var(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host(), prec="double")

    def test_std_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64(
                [np.nan, -1.5, 2.5, np.nan, float("inf"), -float("inf"), 3.0]
            ),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.std(arr)
                i += 1

        expected = np.array([np.std(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_min_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64(
                [np.nan, -1.5, 2.5, np.nan, float("inf"), -float("inf"), 3.0]
            ),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.min(arr)
                i += 1

        expected = np.array([np.min(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_max_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64(
                [np.nan, -1.5, 2.5, np.nan, float("inf"), -float("inf"), 3.0]
            ),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.max(arr)
                i += 1

        expected = np.array([np.max(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_nanmin_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, np.nan]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.nanmin(arr)
                i += 1

        expected = np.array([np.nanmin(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_nanmax_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, np.nan]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.nanmax(arr)
                i += 1

        expected = np.array([np.nanmax(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_nanmean_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, np.nan]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64(
                [np.nan, -1.5, 2.5, np.nan, float("inf"), -float("inf"), 3.0]
            ),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.nanmean(arr)
                i += 1

        expected = np.array([np.nanmean(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_nansum_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, np.nan]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.nansum(arr)
                i += 1

        expected = np.array([np.nansum(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())

    def test_nanprod_basic(self):
        arrays = (
            np.float64([1.0, 2.0, 0.0, -0.0, 1.0, -1.5]),
            np.float64([-0.0, -1.5]),
            np.float64([-1.5, 2.5, np.nan]),
            np.float64([-1.5, 2.5, float("inf")]),
            np.float64([-1.5, 2.5, -float("inf")]),
            np.float64([-1.5, 2.5, float("inf"), -float("inf")]),
            np.float64([np.nan, -1.5, 2.5, np.nan, 3.0]),
            np.float64([5.0, np.nan, -1.5, np.nan]),
            np.float64([np.nan, np.nan]),
        )

        @cuda.jit
        def kernel(out):
            i = 0
            for arr in literal_unroll(arrays):
                out[i] = np.nanprod(arr)
                i += 1

        expected = np.array([np.nanprod(a) for a in arrays], dtype=np.float64)
        out = cuda.to_device(np.zeros(len(arrays), dtype=np.float64))
        kernel[1, 1](out)
        self.assertPreciseEqual(expected, out.copy_to_host())



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np

from numba import cuda
from numba.cuda import float64
from numba.cuda.testing import unittest, CUDATestCase, skip_on_cudasim


def builtin_max(A, B, C):
    i = cuda.grid(1)

    if i >= len(C):
        return

    C[i] = float64(max(A[i], B[i]))


def builtin_min(A, B, C):
    i = cuda.grid(1)

    if i >= len(C):
        return

    C[i] = float64(min(A[i], B[i]))


@skip_on_cudasim("Tests PTX emission")
class TestCudaMinMax(CUDATestCase):
    def _run(
        self,
        kernel,
        numpy_equivalent,
        ptx_instruction,
        dtype_left,
        dtype_right,
        n=5,
    ):
        kernel = cuda.jit(kernel)

        c = np.zeros(n, dtype=np.float64)
        a = np.arange(n, dtype=dtype_left) + 0.5
        b = np.full(n, fill_value=2, dtype=dtype_right)

        kernel[1, c.shape](a, b, c)
        np.testing.assert_allclose(c, numpy_equivalent(a, b))

        ptx = next(p for p in kernel.inspect_asm().values())
        self.assertIn(ptx_instruction, ptx)

    def test_max_f8f8(self):
        self._run(builtin_max, np.maximum, "max.f64", np.float64, np.float64)

    def test_max_f4f8(self):
        self._run(builtin_max, np.maximum, "max.f64", np.float32, np.float64)

    def test_max_f8f4(self):
        self._run(builtin_max, np.maximum, "max.f64", np.float64, np.float32)

    def test_max_f4f4(self):
        self._run(builtin_max, np.maximum, "max.f32", np.float32, np.float32)

    def test_min_f8f8(self):
        self._run(builtin_min, np.minimum, "min.f64", np.float64, np.float64)

    def test_min_f4f8(self):
        self._run(builtin_min, np.minimum, "min.f64", np.float32, np.float64)

    def test_min_f8f4(self):
        self._run(builtin_min, np.minimum, "min.f64", np.float64, np.float32)

    def test_min_f4f4(self):
        self._run(builtin_min, np.minimum, "min.f32", np.float32, np.float32)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import os
from numba import cuda
from numba.cuda.cudadrv.linkable_code import LinkableCode
from numba.cuda.testing import CUDATestCase, skip_on_cudasim
import unittest

TEST_BIN_DIR = os.getenv("NUMBA_CUDA_TEST_BIN_DIR")
if TEST_BIN_DIR:
    test_device_functions_a = os.path.join(
        TEST_BIN_DIR, "test_device_functions.a"
    )
    test_device_functions_cubin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cubin"
    )
    test_device_functions_cu = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cu"
    )
    test_device_functions_fatbin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.fatbin"
    )
    test_device_functions_fatbin_multi = os.path.join(
        TEST_BIN_DIR, "test_device_functions_multi.fatbin"
    )
    test_device_functions_o = os.path.join(
        TEST_BIN_DIR, "test_device_functions.o"
    )
    test_device_functions_ptx = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ptx"
    )
    test_device_functions_ltoir = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ltoir"
    )


class TestLinkableCode(CUDATestCase):
    @skip_on_cudasim(reason="Simulator does not support linkable code")
    @unittest.skipIf(not TEST_BIN_DIR, "necessary binaries not generated.")
    def test_linkable_code_from_path_or_obj(self):
        files_kind = [
            (test_device_functions_a, cuda.Archive),
            (test_device_functions_cubin, cuda.Cubin),
            (test_device_functions_cu, cuda.CUSource),
            (test_device_functions_fatbin, cuda.Fatbin),
            (test_device_functions_o, cuda.Object),
            (test_device_functions_ptx, cuda.PTXSource),
            (test_device_functions_ltoir, cuda.LTOIR),
        ]

        for path, kind in files_kind:
            obj = LinkableCode.from_path_or_obj(path)
            assert isinstance(obj, kind)

            # test identity of from_path_or_obj
            obj2 = LinkableCode.from_path_or_obj(obj)
            assert obj2 is obj



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from contextlib import contextmanager

import numpy as np

from numba import cuda
from numba.cuda.testing import (
    unittest,
    skip_on_cudasim,
    skip_if_external_memmgr,
    CUDATestCase,
)
from numba.cuda.tests.support import captured_stderr
from numba.cuda.core import config


@skip_on_cudasim("not supported on CUDASIM")
class TestDeallocation(CUDATestCase):
    @skip_if_external_memmgr("Deallocation specific to Numba memory management")
    def test_max_pending_count(self):
        # get deallocation manager and flush it
        deallocs = cuda.current_context().memory_manager.deallocations
        deallocs.clear()
        self.assertEqual(len(deallocs), 0)
        # deallocate to maximum count
        for i in range(config.CUDA_DEALLOCS_COUNT):
            cuda.to_device(np.arange(1))
            self.assertEqual(len(deallocs), i + 1)
        # one more to trigger .clear()
        cuda.to_device(np.arange(1))
        self.assertEqual(len(deallocs), 0)

    @skip_if_external_memmgr("Deallocation specific to Numba memory management")
    def test_max_pending_bytes(self):
        # get deallocation manager and flush it
        ctx = cuda.current_context()
        deallocs = ctx.memory_manager.deallocations
        deallocs.clear()
        self.assertEqual(len(deallocs), 0)

        mi = ctx.get_memory_info()

        max_pending = 10**6  # 1MB
        old_ratio = config.CUDA_DEALLOCS_RATIO
        try:
            # change to a smaller ratio
            config.CUDA_DEALLOCS_RATIO = max_pending / mi.total
            # due to round off error (floor is used in calculating
            # _max_pending_bytes) it can be off by 1.
            self.assertAlmostEqual(
                deallocs._max_pending_bytes, max_pending, delta=1
            )

            # allocate half the max size
            # this will not trigger deallocation
            cuda.to_device(np.ones(max_pending // 2, dtype=np.int8))
            self.assertEqual(len(deallocs), 1)

            # allocate another remaining
            # this will not trigger deallocation
            cuda.to_device(
                np.ones(
                    deallocs._max_pending_bytes - deallocs._size, dtype=np.int8
                )
            )
            self.assertEqual(len(deallocs), 2)

            # another byte to trigger .clear()
            cuda.to_device(np.ones(1, dtype=np.int8))
            self.assertEqual(len(deallocs), 0)
        finally:
            # restore old ratio
            config.CUDA_DEALLOCS_RATIO = old_ratio

    @skip_if_external_memmgr("Deallocation specific to Numba memory management")
    def test_defer_cleanup(self):
        harr = np.arange(5)
        darr1 = cuda.to_device(harr)
        deallocs = cuda.current_context().memory_manager.deallocations
        deallocs.clear()
        self.assertEqual(len(deallocs), 0)
        with cuda.defer_cleanup():
            darr2 = cuda.to_device(harr)
            del darr1
            self.assertEqual(len(deallocs), 1)
            del darr2
            self.assertEqual(len(deallocs), 2)
            deallocs.clear()
            self.assertEqual(len(deallocs), 2)

        deallocs.clear()
        self.assertEqual(len(deallocs), 0)

    @skip_if_external_memmgr("Deallocation specific to Numba memory management")
    def test_nested_defer_cleanup(self):
        harr = np.arange(5)
        darr1 = cuda.to_device(harr)
        deallocs = cuda.current_context().memory_manager.deallocations
        deallocs.clear()
        self.assertEqual(len(deallocs), 0)
        with cuda.defer_cleanup():
            with cuda.defer_cleanup():
                darr2 = cuda.to_device(harr)
                del darr1
                self.assertEqual(len(deallocs), 1)
                del darr2
                self.assertEqual(len(deallocs), 2)
                deallocs.clear()
                self.assertEqual(len(deallocs), 2)
            deallocs.clear()
            self.assertEqual(len(deallocs), 2)

        deallocs.clear()
        self.assertEqual(len(deallocs), 0)

    @skip_if_external_memmgr("Deallocation specific to Numba memory management")
    def test_exception(self):
        harr = np.arange(5)
        darr1 = cuda.to_device(harr)
        deallocs = cuda.current_context().memory_manager.deallocations
        deallocs.clear()
        self.assertEqual(len(deallocs), 0)

        class CustomError(Exception):
            pass

        with self.assertRaises(CustomError):
            with cuda.defer_cleanup():
                darr2 = cuda.to_device(harr)
                del darr2
                self.assertEqual(len(deallocs), 1)
                deallocs.clear()
                self.assertEqual(len(deallocs), 1)
                raise CustomError
        deallocs.clear()
        self.assertEqual(len(deallocs), 0)
        del darr1
        self.assertEqual(len(deallocs), 1)
        deallocs.clear()
        self.assertEqual(len(deallocs), 0)

    @contextmanager
    def check_ignored_exception(self, ctx):
        with captured_stderr() as cap:
            yield
            ctx.deallocations.clear()
        self.assertFalse(cap.getvalue())

    def test_del_stream(self):
        ctx = cuda.current_context()
        stream = ctx.create_stream()
        with self.check_ignored_exception(ctx):
            del stream

    def test_del_event(self):
        ctx = cuda.current_context()
        event = ctx.create_event()
        with self.check_ignored_exception(ctx):
            del event

    def test_del_pinned_memory(self):
        ctx = cuda.current_context()
        mem = ctx.memhostalloc(32)
        with self.check_ignored_exception(ctx):
            del mem

    def test_del_mapped_memory(self):
        ctx = cuda.current_context()
        mem = ctx.memhostalloc(32, mapped=True)
        with self.check_ignored_exception(ctx):
            del mem

    def test_del_device_memory(self):
        ctx = cuda.current_context()
        mem = ctx.memalloc(32)
        with self.check_ignored_exception(ctx):
            del mem

    def test_del_managed_memory(self):
        ctx = cuda.current_context()
        mem = ctx.memallocmanaged(32)
        with self.check_ignored_exception(ctx):
            del mem

    def test_del_pinned_contextmanager(self):
        # Check that temporarily pinned memory is unregistered immediately,
        # such that it can be re-pinned at any time
        class PinnedException(Exception):
            pass

        arr = np.zeros(1)
        ctx = cuda.current_context()
        ctx.deallocations.clear()
        with self.check_ignored_exception(ctx):
            with cuda.pinned(arr):
                pass
            with cuda.pinned(arr):
                pass
            # Should also work inside a `defer_cleanup` block
            with cuda.defer_cleanup():
                with cuda.pinned(arr):
                    pass
                with cuda.pinned(arr):
                    pass
            # Should also work when breaking out of the block due to an
            # exception
            try:
                with cuda.pinned(arr):
                    raise PinnedException
            except PinnedException:
                with cuda.pinned(arr):
                    pass

    def test_del_mapped_contextmanager(self):
        # Check that temporarily mapped memory is unregistered immediately,
        # such that it can be re-mapped at any time
        class MappedException(Exception):
            pass

        arr = np.zeros(1)
        ctx = cuda.current_context()
        ctx.deallocations.clear()
        with self.check_ignored_exception(ctx):
            with cuda.mapped(arr):
                pass
            with cuda.mapped(arr):
                pass
            # Should also work inside a `defer_cleanup` block
            with cuda.defer_cleanup():
                with cuda.mapped(arr):
                    pass
                with cuda.mapped(arr):
                    pass
            # Should also work when breaking out of the block due to an
            # exception
            try:
                with cuda.mapped(arr):
                    raise MappedException
            except MappedException:
                with cuda.mapped(arr):
                    pass


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import os
from math import sqrt
from numba import cuda
from numba.cuda import float32, int16, int32, int64, types, uint32, void
from numba.cuda import (
    compile,
    compile_for_current_device,
    compile_ptx,
    compile_ptx_for_current_device,
    compile_all,
    LinkableCode,
)
from numba.cuda.testing import skip_on_cudasim, unittest, CUDATestCase

TEST_BIN_DIR = os.getenv("NUMBA_CUDA_TEST_BIN_DIR")
if TEST_BIN_DIR:
    test_device_functions_a = os.path.join(
        TEST_BIN_DIR, "test_device_functions.a"
    )
    test_device_functions_cubin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cubin"
    )
    test_device_functions_cu = os.path.join(
        TEST_BIN_DIR, "test_device_functions.cu"
    )
    test_device_functions_fatbin = os.path.join(
        TEST_BIN_DIR, "test_device_functions.fatbin"
    )
    test_device_functions_fatbin_multi = os.path.join(
        TEST_BIN_DIR, "test_device_functions_multi.fatbin"
    )
    test_device_functions_o = os.path.join(
        TEST_BIN_DIR, "test_device_functions.o"
    )
    test_device_functions_ptx = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ptx"
    )
    test_device_functions_ltoir = os.path.join(
        TEST_BIN_DIR, "test_device_functions.ltoir"
    )


# A test function at the module scope to ensure we get the name right for the C
# ABI whether a function is at module or local scope.
def f_module(x, y):
    return x + y


@skip_on_cudasim("Compilation unsupported in the simulator")
class TestCompile(unittest.TestCase):
    def _handle_compile_result(self, ret, compile_function):
        ptx_or_code_list, resty = ret
        if compile_function in (compile_ptx, compile):
            ptx = ptx_or_code_list
        else:
            ptx = ptx_or_code_list[0]
        return ptx, resty

    def test_global_kernel(self):
        with self.subTest("compile_ptx"):
            self._test_global_kernel(compile_ptx, {})

        with self.subTest("compile_all"):
            self._test_global_kernel(
                compile_all, {"device": False, "abi": "numba", "output": "ptx"}
            )

    def _test_global_kernel(self, compile_function, default_kwargs):
        def f(r, x, y):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = x[i] + y[i]

        args = (float32[:], float32[:], float32[:])

        ret = compile_function(f, args, **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)

        # Kernels should not have a func_retval parameter
        self.assertNotIn("func_retval", ptx)
        # .visible .func is used to denote a device function
        self.assertNotIn(".visible .func", ptx)
        # .visible .entry would denote the presence of a global function
        self.assertIn(".visible .entry", ptx)
        # Return type for kernels should always be void
        self.assertEqual(resty, void)

    def test_device_function(self):
        with self.subTest("compile_ptx"):
            self._test_device_function(compile_ptx, {"device": True})

        with self.subTest("compile_all"):
            self._test_device_function(
                compile_all, {"device": True, "abi": "c", "output": "ptx"}
            )

    def _test_device_function(self, compile_function, default_kwargs):
        def add(x, y):
            return x + y

        args = (float32, float32)

        ret = compile_function(add, args, **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)

        # Device functions take a func_retval parameter for storing the
        # returned value in by reference
        self.assertIn("func_retval", ptx)
        # .visible .func is used to denote a device function
        self.assertIn(".visible .func", ptx)
        # .visible .entry would denote the presence of a global function
        self.assertNotIn(".visible .entry", ptx)
        # Inferred return type as expected?
        self.assertEqual(resty, float32)

        # Check that function's output matches signature
        sig_int32 = int32(int32, int32)
        ret = compile_function(add, sig_int32, **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.assertEqual(resty, int32)

        sig_int16 = int16(int16, int16)
        ret = compile_function(add, sig_int16, **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.assertEqual(resty, int16)
        # Using string as signature
        sig_string = "uint32(uint32, uint32)"
        ret = compile_function(add, sig_string, **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.assertEqual(resty, uint32)

    def test_fastmath(self):
        with self.subTest("compile_ptx"):
            self._test_fastmath(compile_ptx, {"device": True})

        with self.subTest("compile_all"):
            self._test_fastmath(compile_all, {"device": True, "output": "ptx"})

    def _test_fastmath(self, compile_function, default_kwargs):
        def f(x, y, z, d):
            return sqrt((x * y + z) / d)

        args = (float32, float32, float32, float32)

        # Without fastmath, fma contraction is enabled by default, but ftz and
        # approximate div / sqrt are not.
        ret = compile_function(f, args, **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.assertIn("fma.rn.f32", ptx)
        self.assertIn("div.rn.f32", ptx)
        self.assertIn("sqrt.rn.f32", ptx)

        # With fastmath, ftz and approximate div / sqrt are enabled
        ret = compile_function(f, args, fastmath=True, **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.assertIn("fma.rn.ftz.f32", ptx)
        self.assertIn("div.approx.ftz.f32", ptx)
        self.assertIn("sqrt.approx.ftz.f32", ptx)

    def check_debug_info(self, ptx):
        # A debug_info section should exist in the PTX. Whitespace varies
        # between CUDA toolkit versions.
        self.assertRegex(ptx, "\\.section\\s+\\.debug_info")
        # A .file directive should be produced and include the name of the
        # source. The path and whitespace may vary, so we accept anything
        # ending in the filename of this module.
        self.assertRegex(ptx, '\\.file.*test_compiler.py"')

    def test_device_function_with_debug(self):
        # See Issue #6719 - this ensures that compilation with debug succeeds
        # with CUDA 11.2 / NVVM 7.0 onwards. Previously it failed because NVVM
        # IR version metadata was not added when compiling device functions,
        # and NVVM assumed DBG version 1.0 if not specified, which is
        # incompatible with the 3.0 IR we use. This was specified only for
        # kernels.

        with self.subTest("compile_ptx"):
            self._test_device_function_with_debug(
                compile_ptx, {"device": True, "debug": True, "opt": False}
            )

        with self.subTest("compile_all"):
            self._test_device_function_with_debug(
                compile_all,
                {
                    "device": True,
                    "debug": True,
                    "opt": False,
                    "output": "ptx",
                },
            )

    def _test_device_function_with_debug(
        self, compile_function, default_kwargs
    ):
        def f():
            pass

        ret = compile_function(f, (), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.check_debug_info(ptx)

    def test_kernel_with_debug(self):
        # Inspired by (but not originally affected by) Issue #6719

        with self.subTest("compile_ptx"):
            self._test_kernel_with_debug(
                compile_ptx, {"debug": True, "opt": False}
            )

        with self.subTest("compile_all"):
            self._test_kernel_with_debug(
                compile_all,
                {
                    "device": False,
                    "abi": "numba",
                    "debug": True,
                    "opt": False,
                    "output": "ptx",
                },
            )

    def _test_kernel_with_debug(self, compile_function, default_kwargs):
        def f():
            pass

        ret = compile_function(f, (), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.check_debug_info(ptx)

    def check_line_info(self, ptx):
        # A .file directive should be produced and include the name of the
        # source. The path and whitespace may vary, so we accept anything
        # ending in the filename of this module.
        self.assertRegex(ptx, '\\.file.*test_compiler.py"')

    def test_device_function_with_line_info(self):
        with self.subTest("compile_ptx"):
            self._test_device_function_with_line_info(
                compile_ptx, {"device": True, "lineinfo": True}
            )

        with self.subTest("compile_all"):
            self._test_device_function_with_line_info(
                compile_all,
                {
                    "device": True,
                    "abi": "numba",
                    "lineinfo": True,
                    "output": "ptx",
                },
            )

    def _test_device_function_with_line_info(
        self, compile_function, default_kwargs
    ):
        def f():
            pass

        ret = compile_function(f, (), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.check_line_info(ptx)

    def test_kernel_with_line_info(self):
        with self.subTest("compile_ptx"):
            self._test_kernel_with_line_info(compile_ptx, {"lineinfo": True})

        with self.subTest("compile_all"):
            self._test_kernel_with_line_info(
                compile_all,
                {
                    "device": False,
                    "abi": "numba",
                    "lineinfo": True,
                    "output": "ptx",
                },
            )

    def _test_kernel_with_line_info(self, compile_function, default_kwargs):
        def f():
            pass

        ret = compile_function(f, (), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.check_line_info(ptx)

    def test_non_void_return_type(self):
        def f(x, y):
            return x[0] + y[0]

        with self.subTest("compile_ptx"):
            with self.assertRaisesRegex(
                TypeError, "must have void return type"
            ):
                compile_ptx(f, (uint32[::1], uint32[::1]))

        with self.subTest("compile_all"):
            with self.assertRaisesRegex(
                TypeError, "must have void return type"
            ):
                compile_all(
                    f,
                    (uint32[::1], uint32[::1]),
                    device=False,
                    abi="numba",
                    output="ptx",
                )

    def test_c_abi_disallowed_for_kernel(self):
        def f(x, y):
            return x + y

        with self.subTest("compile_ptx"):
            with self.assertRaisesRegex(
                NotImplementedError, "The C ABI is not supported for kernels"
            ):
                compile_ptx(f, (int32, int32), abi="c")

        with self.subTest("compile_all"):
            with self.assertRaisesRegex(
                NotImplementedError, "The C ABI is not supported for kernels"
            ):
                compile_all(
                    f, (int32, int32), abi="c", device=False, output="ptx"
                )

    def test_unsupported_abi(self):
        def f(x, y):
            return x + y

        with self.subTest("compile_ptx"):
            with self.assertRaisesRegex(
                NotImplementedError, "Unsupported ABI: fastcall"
            ):
                compile_ptx(f, (int32, int32), abi="fastcall")

        with self.subTest("compile_all"):
            with self.assertRaisesRegex(
                NotImplementedError, "Unsupported ABI: fastcall"
            ):
                compile_all(f, (int32, int32), abi="fastcall", output="ptx")

    def test_c_abi_device_function(self):
        with self.subTest("compile_ptx"):
            self._test_c_abi_device_function(
                compile_ptx, {"device": True, "abi": "c"}
            )

        with self.subTest("compile_all"):
            self._test_c_abi_device_function(
                compile_all, {"device": True, "abi": "c", "output": "ptx"}
            )

    def _test_c_abi_device_function(self, compile_function, default_kwargs):
        def f(x, y):
            return x + y

        # 32-bit signature
        ret = compile_function(f, int32(int32, int32), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        # There should be no more than two parameters
        self.assertNotIn(ptx, "param_2")
        # The function name should match the Python function name (not the
        # qualname, which includes additional info), and its return value
        # should be 32 bits
        self.assertRegex(
            ptx,
            r"\.visible\s+\.func\s+\(\.param\s+\.b32\s+"
            r"func_retval0\)\s+f\(",
        )

        # 64-bit signature should produce 64-bit return parameter
        ret = compile_function(f, int64(int64, int64), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)
        self.assertRegex(ptx, r"\.visible\s+\.func\s+\(\.param\s+\.b64")

    def test_c_abi_device_function_module_scope(self):
        with self.subTest("compile_ptx"):
            self._test_c_abi_device_function_module_scope(
                compile_ptx, {"device": True, "abi": "c"}
            )

        with self.subTest("compile_all"):
            self._test_c_abi_device_function_module_scope(
                compile_all,
                {"device": True, "abi": "c", "output": "ptx"},
            )

    def _test_c_abi_device_function_module_scope(
        self, compile_function, default_kwargs
    ):
        ret = compile_function(f_module, int32(int32, int32), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)

        # The function name should match the Python function name, and its
        # return value should be 32 bits
        self.assertRegex(
            ptx,
            r"\.visible\s+\.func\s+\(\.param\s+\.b32\s+"
            r"func_retval0\)\s+f_module\(",
        )

    def test_c_abi_with_abi_name(self):
        abi_info = {"abi_name": "_Z4funcii"}

        with self.subTest("compile_ptx"):
            self._test_c_abi_with_abi_name(
                compile_ptx,
                {"device": True, "abi": "c", "abi_info": abi_info},
            )

        with self.subTest("compile_all"):
            self._test_c_abi_with_abi_name(
                compile_all,
                {
                    "device": True,
                    "abi": "c",
                    "abi_info": abi_info,
                    "output": "ptx",
                },
            )

    def _test_c_abi_with_abi_name(self, compile_function, default_kwargs):
        ret = compile_function(f_module, int32(int32, int32), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)

        # The function name should match the one given in the ABI info, and its
        # return value should be 32 bits
        self.assertRegex(
            ptx,
            r"\.visible\s+\.func\s+\(\.param\s+\.b32\s+"
            r"func_retval0\)\s+_Z4funcii\(",
        )

    def test_compile_defaults_to_c_abi(self):
        with self.subTest("compile"):
            self._test_compile_defaults_to_c_abi(compile, {"device": True})

        with self.subTest("compile_all"):
            self._test_compile_defaults_to_c_abi(
                compile_all,
                {"device": True, "output": "ptx"},
            )

    def _test_compile_defaults_to_c_abi(self, compile_function, default_kwargs):
        ret = compile_function(f_module, int32(int32, int32), **default_kwargs)
        ptx, resty = self._handle_compile_result(ret, compile_function)

        # The function name should match the Python function name, and its
        # return value should be 32 bits
        self.assertRegex(
            ptx,
            r"\.visible\s+\.func\s+\(\.param\s+\.b32\s+"
            r"func_retval0\)\s+f_module\(",
        )

    def test_compile_to_ltoir(self):
        with self.subTest("compile"):
            self._test_compile_to_ltoir(
                compile, {"device": True, "output": "ltoir"}
            )

        with self.subTest("compile_all"):
            self._test_compile_to_ltoir(
                compile_all,
                {"device": True, "abi": "c", "output": "ltoir"},
            )

    def _test_compile_to_ltoir(self, compile_function, default_kwargs):
        ret = compile_function(f_module, int32(int32, int32), **default_kwargs)
        code, resty = self._handle_compile_result(ret, compile_function)

        # There are no tools to interpret the LTOIR output, but we can check
        # that we appear to have obtained an LTOIR file. This magic number is
        # not documented, but is expected to remain consistent.
        LTOIR_MAGIC = 0x7F4E43ED
        header = int.from_bytes(code[:4], byteorder="little")
        self.assertEqual(header, LTOIR_MAGIC)
        self.assertEqual(resty, int32)

    def test_compile_to_invalid_error(self):
        illegal_output = "illegal"
        msg = f"Unsupported output type: {illegal_output}"
        with self.subTest("compile"):
            with self.assertRaisesRegex(NotImplementedError, msg):
                compile(
                    f_module,
                    int32(int32, int32),
                    device=True,
                    output=illegal_output,
                )

        with self.subTest("compile_all"):
            with self.assertRaisesRegex(NotImplementedError, msg):
                compile_all(
                    f_module,
                    int32(int32, int32),
                    device=True,
                    abi="c",
                    output=illegal_output,
                )

    def test_functioncompiler_locals(self):
        # Tests against regression fixed in:
        # https://github.com/NVIDIA/numba-cuda/pull/381
        #
        # "AttributeError: '_FunctionCompiler' object has no attribute
        # 'locals'"
        cond = None

        @cuda.jit("void(float32[::1])")
        def f(b_arg):
            b_smem = cuda.shared.array(shape=(1,), dtype=float32)

            if cond:
                b_smem[0] = b_arg[0]

    @unittest.skipIf(not TEST_BIN_DIR, "necessary binaries not generated.")
    def test_compile_all_with_external_functions(self):
        for link in [
            test_device_functions_a,
            test_device_functions_cubin,
            test_device_functions_cu,
            test_device_functions_fatbin,
            test_device_functions_fatbin_multi,
            test_device_functions_o,
            test_device_functions_ptx,
            test_device_functions_ltoir,
        ]:
            with self.subTest(link=link):
                add = cuda.declare_device(
                    "add_from_numba", "uint32(uint32, uint32)", link=[link]
                )

                def f(z, x, y):
                    z[0] = add(x, y)

                code_list, resty = compile_all(
                    f, (uint32[::1], uint32, uint32), device=False, abi="numba"
                )

                assert resty == void
                assert len(code_list) == 2
                link_obj = LinkableCode.from_path(link)
                if link_obj.kind == "cu":
                    # if link is a cu file, result contains a compiled object code
                    from cuda.core.experimental import ObjectCode

                    assert isinstance(code_list[1], ObjectCode)
                else:
                    assert code_list[1].kind == link_obj.kind

    @unittest.skipIf(not TEST_BIN_DIR, "necessary binaries not generated.")
    def test_compile_all_lineinfo(self):
        add = cuda.declare_device(
            "add", "float32(float32, float32)", link=[test_device_functions_cu]
        )

        def f(z, x, y):
            z[0] = add(x, y)

        args = (float32[::1], float32, float32)
        code_list, resty = compile_all(
            f, args, lineinfo=True, output="ptx", device=False, abi="numba"
        )
        assert len(code_list) == 2

        self.assertRegex(
            str(code_list[1].code.decode()),
            r"\.file.*test_device_functions",
        )

    @unittest.skipIf(not TEST_BIN_DIR, "necessary binaries not generated.")
    def test_compile_all_debug(self):
        add = cuda.declare_device(
            "add", "float32(float32, float32)", link=[test_device_functions_cu]
        )

        def f(z, x, y):
            z[0] = add(x, y)

        args = (float32[::1], float32, float32)
        code_list, resty = compile_all(
            f,
            args,
            debug=True,
            output="ptx",
            device=False,
            abi="numba",
            opt=False,
        )
        assert len(code_list) == 2

        self.assertRegex(
            str(code_list[1].code.decode()), r"\.section\s+\.debug_info"
        )


@skip_on_cudasim("Compilation unsupported in the simulator")
class TestCompileForCurrentDevice(CUDATestCase):
    def _check_ptx_for_current_device(self, compile_function):
        def add(x, y):
            return x + y

        args = (float32, float32)
        ptx, resty = compile_function(add, args, device=True)

        # Check we target the current device's compute capability, or the
        # closest compute capability supported by the current toolkit.
        device_cc = cuda.get_current_device().compute_capability
        cc = cuda.cudadrv.nvrtc.find_closest_arch(device_cc)
        target = f".target sm_{cc[0]}{cc[1]}"
        self.assertIn(target, ptx)

    def test_compile_ptx_for_current_device(self):
        self._check_ptx_for_current_device(compile_ptx_for_current_device)

    def test_compile_for_current_device(self):
        self._check_ptx_for_current_device(compile_for_current_device)


@skip_on_cudasim("Compilation unsupported in the simulator")
class TestCompileOnlyTests(unittest.TestCase):
    """For tests where we can only check correctness by examining the compiler
    output rather than observing the effects of execution."""

    def test_nanosleep(self):
        def use_nanosleep(x):
            # Sleep for a constant time
            cuda.nanosleep(32)
            # Sleep for a variable time
            cuda.nanosleep(x)

        ptx, resty = compile_ptx(use_nanosleep, (uint32,))

        nanosleep_count = 0
        for line in ptx.split("\n"):
            if "nanosleep.u32" in line:
                nanosleep_count += 1

        expected = 2
        self.assertEqual(
            expected,
            nanosleep_count,
            (
                f"Got {nanosleep_count} nanosleep instructions, "
                f"expected {expected}"
            ),
        )


@skip_on_cudasim("Compilation unsupported in the simulator")
class TestCompileWithLaunchBounds(unittest.TestCase):
    def _test_launch_bounds_common(self, launch_bounds):
        def f():
            pass

        sig = "void()"
        ptx, resty = cuda.compile_ptx(f, sig, launch_bounds=launch_bounds)
        self.assertIsInstance(resty, types.NoneType)
        # Match either `.maxntid, 128, 1, 1` or `.maxntid 128` on a line by
        # itself:
        self.assertRegex(ptx, r".maxntid\s+128(?:,\s+1,\s+1)?\s*\n")
        return ptx

    def test_launch_bounds_scalar(self):
        launch_bounds = 128
        ptx = self._test_launch_bounds_common(launch_bounds)

        self.assertNotIn(".minnctapersm", ptx)
        self.assertNotIn(".maxclusterrank", ptx)

    def test_launch_bounds_tuple(self):
        launch_bounds = (128,)
        ptx = self._test_launch_bounds_common(launch_bounds)

        self.assertNotIn(".minnctapersm", ptx)
        self.assertNotIn(".maxclusterrank", ptx)

    def test_launch_bounds_with_min_cta(self):
        launch_bounds = (128, 2)
        ptx = self._test_launch_bounds_common(launch_bounds)

        self.assertRegex(ptx, r".minnctapersm\s+2")
        self.assertNotIn(".maxclusterrank", ptx)

    def test_launch_bounds_with_max_cluster_rank(self):
        def f():
            pass

        launch_bounds = (128, 2, 4)
        cc = (9, 0)
        sig = "void()"
        ptx, resty = cuda.compile_ptx(
            f, sig, launch_bounds=launch_bounds, cc=cc
        )
        self.assertIsInstance(resty, types.NoneType)
        self.assertRegex(ptx, r".maxntid\s+128,\s+1,\s+1")

        self.assertRegex(ptx, r".minnctapersm\s+2")
        self.assertRegex(ptx, r".maxclusterrank\s+4")

    def test_too_many_launch_bounds(self):
        def f():
            pass

        sig = "void()"
        launch_bounds = (128, 2, 4, 8)

        with self.assertRaisesRegex(ValueError, "Got 4 launch bounds:"):
            cuda.compile_ptx(f, sig, launch_bounds=launch_bounds)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba import cuda
from numba.cuda import float32, int32
from numba.cuda.core.errors import NumbaInvalidConfigWarning
import pytest

from numba.cuda.testing import CUDATestCase, skip_on_cudasim
import re
import unittest


@skip_on_cudasim("Simulator does not produce lineinfo")
class TestCudaLineInfo(CUDATestCase):
    def _loc_directive_regex(self):
        # This is used in several tests

        pat = (
            r"\.loc"  # .loc directive beginning
            r"\s+[0-9]+"  # whitespace then file index
            r"\s+[0-9]+"  # whitespace then line number
            r"\s+[0-9]+"  # whitespace then column position
        )
        return re.compile(pat)

    def _check(self, fn, sig, expect):
        fn.compile(sig)
        llvm = fn.inspect_llvm(sig)
        ptx = fn.inspect_asm(sig)
        assertfn = self.assertIsNotNone if expect else self.assertIsNone

        # DICompileUnit debug info metadata should all be of the
        # DebugDirectivesOnly kind, and not the FullDebug kind
        pat = (
            r"!DICompileUnit\(.*"  # Opening of DICompileUnit metadata. Since
            # the order of attributes is not
            # guaranteed, we need to match arbitrarily
            # afterwards.
            r"emissionKind:\s+"  # The emissionKind attribute followed by
            # whitespace.
            r"DebugDirectivesOnly"  # The correct emissionKind.
        )
        match = re.compile(pat).search(llvm)
        assertfn(match, msg=ptx)

        pat = (
            r"!DICompileUnit\(.*"  # Same as the pattern above, but for the
            r"emissionKind:\s+"  # incorrect FullDebug emissionKind.
            r"FullDebug"  #
        )
        match = re.compile(pat).search(llvm)
        self.assertIsNone(match, msg=ptx)

        # The name of this file should be present in the line mapping
        # if lineinfo was propagated through correctly.
        pat = (
            r"\.file"  # .file directive beginning
            r"\s+[0-9]+\s+"  # file number surrounded by whitespace
            r'".*test_lineinfo.py"'  # filename in quotes, ignoring full path
        )
        match = re.compile(pat).search(ptx)
        assertfn(match, msg=ptx)

        # .loc directives should be present in the ptx
        self._loc_directive_regex().search(ptx)
        assertfn(match, msg=ptx)

        # Debug info sections should not be present when only lineinfo is
        # generated
        pat = (
            r"\.section\s+"  # .section directive beginning
            r"\.debug_info"  # Section named ".debug_info"
        )
        match = re.compile(pat).search(ptx)
        self.assertIsNone(match, msg=ptx)

    def test_no_lineinfo_in_asm(self):
        @cuda.jit(lineinfo=False)
        def foo(x):
            x[0] = 1

        self._check(foo, sig=(int32[:],), expect=False)

    def test_lineinfo_in_asm(self):
        @cuda.jit(lineinfo=True)
        def foo(x):
            x[0] = 1

        self._check(foo, sig=(int32[:],), expect=True)

    def test_lineinfo_maintains_error_model(self):
        sig = (float32[::1], float32[::1])

        @cuda.jit(sig, lineinfo=True)
        def divide_kernel(x, y):
            x[0] /= y[0]

        llvm = divide_kernel.inspect_llvm(sig)

        # When the error model is Python, the device function returns 1 to
        # signal an exception (e.g. divide by zero) has occurred. When the
        # error model is the default NumPy one (as it should be when only
        # lineinfo is enabled) the device function always returns 0.
        self.assertNotIn("ret i32 1", llvm)

    def test_no_lineinfo_in_device_function(self):
        # Ensure that no lineinfo is generated in device functions by default.
        @cuda.jit
        def callee(x):
            x[0] += 1

        @cuda.jit
        def caller(x):
            x[0] = 1
            callee(x)

        sig = (int32[:],)
        self._check(caller, sig=sig, expect=False)

    def test_lineinfo_in_device_function(self):
        # First we define a device function / kernel pair and run the usual
        # checks on the generated LLVM and PTX.

        @cuda.jit(lineinfo=True)
        def callee(x):
            x[0] += 1

        @cuda.jit(lineinfo=True)
        def caller(x):
            x[0] = 1
            callee(x)

        sig = (int32[:],)
        self._check(caller, sig=sig, expect=True)

        # Now we can check the PTX of the device function specifically.

        ptx = caller.inspect_asm(sig)
        ptxlines = ptx.splitlines()

        # Check that there is no device function in the PTX

        # A line beginning with ".weak .func" that identifies a device function
        devfn_start = re.compile(r"^\.weak\s+\.func")

        for line in ptxlines:
            if devfn_start.match(line) is not None:
                self.fail(f"Found device function in PTX:\n\n{ptx}")

        # Scan for .loc directives that refer to an inlined device function

        loc_directive = self._loc_directive_regex()
        found = False

        for line in ptxlines:
            if loc_directive.search(line) is not None:
                if "inlined_at" in line:
                    found = True
                    break

        if not found:
            self.fail(
                f"No .loc directive with inlined_at info foundin:\n\n{ptx}"
            )

        # We also inspect the LLVM to ensure that there's debug info for each
        # subprogram (function). A lightweight way to check this is to ensure
        # that we have as many DISubprograms as we expect.

        llvm = caller.inspect_llvm(sig)
        subprograms = 0
        for line in llvm.splitlines():
            if "distinct !DISubprogram" in line:
                subprograms += 1

        # One DISubprogram for each of:
        # - The caller
        # - The callee
        expected_subprograms = 2

        self.assertEqual(
            subprograms,
            expected_subprograms,
            f'"Expected {expected_subprograms} DISubprograms; '
            f"got {subprograms}",
        )

    def test_debug_and_lineinfo_warning(self):
        with pytest.warns(
            NumbaInvalidConfigWarning,
            match="debug and lineinfo are mutually exclusive",
        ):
            # We pass opt=False to prevent the warning about opt and debug
            # occurring as well
            @cuda.jit(debug=True, lineinfo=True, opt=False)
            def f():
                pass

    def test_lineinfo_with_compile_internal(self):
        # Calling a function implemented using compile_internal should not
        # enable full debug info generation. See Numba-CUDA Issue #271,
        # https://github.com/NVIDIA/numba-cuda/issues/271

        @cuda.jit("void(complex128[::1], complex128[::1])", lineinfo=True)
        def complex_abs_use(r, x):
            r[0] = abs(x[0])

        cc = cuda.get_current_device().compute_capability
        ov = complex_abs_use.overloads[complex_abs_use.signatures[0]]
        ptx = ov.inspect_asm(cc)

        target = ".target sm_%s%s" % cc
        target_debug = f"{target}, debug"
        self.assertIn(target, ptx)
        self.assertNotIn(target_debug, ptx)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np
from numba import cuda
from numba.cuda.testing import CUDATestCase
import unittest
from numba.cuda import config


def reinterpret_array_type(byte_arr, start, stop, output):
    # Tested with just one thread
    val = byte_arr[start:stop].view(np.int32)[0]
    output[0] = val


class TestCudaArrayMethods(CUDATestCase):
    def setUp(self):
        self.old_nrt_setting = config.CUDA_ENABLE_NRT
        config.CUDA_ENABLE_NRT = True
        super(TestCudaArrayMethods, self).setUp()

    def tearDown(self):
        config.CUDA_ENABLE_NRT = self.old_nrt_setting
        super(TestCudaArrayMethods, self).tearDown()

    def test_reinterpret_array_type(self):
        """
        Reinterpret byte array as int32 in the GPU.
        """
        pyfunc = reinterpret_array_type
        kernel = cuda.jit(pyfunc)

        byte_arr = np.arange(256, dtype=np.uint8)
        itemsize = np.dtype(np.int32).itemsize
        for start in range(0, 256, itemsize):
            stop = start + itemsize
            expect = byte_arr[start:stop].view(np.int32)[0]

            output = np.zeros(1, dtype=np.int32)
            kernel[1, 1](byte_arr, start, stop, output)

            got = output[0]
            self.assertEqual(expect, got)

    def test_array_copy(self):
        val = np.array([1, 2, 3])[::-1]

        @cuda.jit
        def kernel(out):
            q = val.copy()
            for i in range(len(out)):
                out[i] = q[i]

        out = cuda.to_device(np.zeros(len(val), dtype="float64"))

        kernel[1, 1](out)
        for i, j in zip(out.copy_to_host(), val):
            self.assertEqual(i, j)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import functools
import warnings
import numpy as np
import unittest
from numba.cuda import HAS_NUMBA

if HAS_NUMBA:
    from numba import njit
from numba import cuda
from numba.cuda import config, types
from numba.cuda.testing import skip_on_standalone_numba_cuda
from numba.cuda.typing.typeof import typeof
from numba.cuda.np import numpy_support
from numba.cuda.tests.support import TestCase


class BaseUFuncTest:
    def setUp(self):
        self.inputs = [
            (np.uint32(0), types.uint32),
            (np.uint32(1), types.uint32),
            (np.int32(-1), types.int32),
            (np.int32(0), types.int32),
            (np.int32(1), types.int32),
            (np.uint64(0), types.uint64),
            (np.uint64(1), types.uint64),
            (np.int64(-1), types.int64),
            (np.int64(0), types.int64),
            (np.int64(1), types.int64),
            (np.float32(-0.5), types.float32),
            (np.float32(0.0), types.float32),
            (np.float32(0.5), types.float32),
            (np.float64(-0.5), types.float64),
            (np.float64(0.0), types.float64),
            (np.float64(0.5), types.float64),
            (np.array([0, 1], dtype="u4"), types.Array(types.uint32, 1, "C")),
            (np.array([0, 1], dtype="u8"), types.Array(types.uint64, 1, "C")),
            (
                np.array([-1, 0, 1], dtype="i4"),
                types.Array(types.int32, 1, "C"),
            ),
            (
                np.array([-1, 0, 1], dtype="i8"),
                types.Array(types.int64, 1, "C"),
            ),
            (
                np.array([-0.5, 0.0, 0.5], dtype="f4"),
                types.Array(types.float32, 1, "C"),
            ),
            (
                np.array([-0.5, 0.0, 0.5], dtype="f8"),
                types.Array(types.float64, 1, "C"),
            ),
            (np.array([0, 1], dtype=np.int8), types.Array(types.int8, 1, "C")),
            (
                np.array([0, 1], dtype=np.int16),
                types.Array(types.int16, 1, "C"),
            ),
            (
                np.array([0, 1], dtype=np.uint8),
                types.Array(types.uint8, 1, "C"),
            ),
            (
                np.array([0, 1], dtype=np.uint16),
                types.Array(types.uint16, 1, "C"),
            ),
        ]

    @skip_on_standalone_numba_cuda
    @functools.lru_cache(maxsize=None)
    def _compile(self, pyfunc, args, nrt=False):
        # NOTE: to test the implementation of Numpy ufuncs, we disable
        # rewriting of array expressions.
        return njit(args, _nrt=nrt, no_rewrites=True)(pyfunc)

    def _determine_output_type(
        self, input_type, int_output_type=None, float_output_type=None
    ):
        ty = input_type
        if isinstance(ty, types.Array):
            ndim = ty.ndim
            ty = ty.dtype
        else:
            ndim = 1

        if ty in types.signed_domain:
            if int_output_type:
                output_type = types.Array(int_output_type, ndim, "C")
            else:
                output_type = types.Array(ty, ndim, "C")
        elif ty in types.unsigned_domain:
            if int_output_type:
                output_type = types.Array(int_output_type, ndim, "C")
            else:
                output_type = types.Array(ty, ndim, "C")
        else:
            if float_output_type:
                output_type = types.Array(float_output_type, ndim, "C")
            else:
                output_type = types.Array(ty, ndim, "C")
        return output_type


class BasicUFuncTest(BaseUFuncTest):
    def _make_ufunc_usecase(self, ufunc):
        return _make_ufunc_usecase(ufunc)

    def basic_ufunc_test(
        self,
        ufunc,
        skip_inputs=[],
        additional_inputs=[],
        int_output_type=None,
        float_output_type=None,
        kinds="ifc",
        positive_only=False,
    ):
        # Necessary to avoid some Numpy warnings being silenced, despite
        # the simplefilter() call below.
        self.reset_module_warnings(__name__)

        pyfunc = self._make_ufunc_usecase(ufunc)

        inputs = list(self.inputs) + additional_inputs

        for input_tuple in inputs:
            input_operand = input_tuple[0]
            input_type = input_tuple[1]

            is_tuple = isinstance(input_operand, tuple)
            if is_tuple:
                args = input_operand
            else:
                args = (input_operand,) * ufunc.nin

            if input_type in skip_inputs:
                continue
            if positive_only and np.any(args[0] < 0):
                continue

            # Some ufuncs don't allow all kinds of arguments
            if args[0].dtype.kind not in kinds:
                continue

            output_type = self._determine_output_type(
                input_type, int_output_type, float_output_type
            )

            input_types = (input_type,) * ufunc.nin
            output_types = (output_type,) * ufunc.nout
            argtys = input_types + output_types
            cfunc = self._compile(pyfunc, argtys)

            if isinstance(args[0], np.ndarray):
                results = [
                    np.zeros(args[0].shape, dtype=out_ty.dtype.name)
                    for out_ty in output_types
                ]
                expected = [
                    np.zeros(args[0].shape, dtype=out_ty.dtype.name)
                    for out_ty in output_types
                ]
            else:
                results = [
                    np.zeros(1, dtype=out_ty.dtype.name)
                    for out_ty in output_types
                ]
                expected = [
                    np.zeros(1, dtype=out_ty.dtype.name)
                    for out_ty in output_types
                ]

            invalid_flag = False
            with warnings.catch_warnings(record=True) as warnlist:
                warnings.simplefilter("always")
                pyfunc(*args, *expected)

                warnmsg = "invalid value encountered"
                for thiswarn in warnlist:
                    if issubclass(thiswarn.category, RuntimeWarning) and str(
                        thiswarn.message
                    ).startswith(warnmsg):
                        invalid_flag = True

            cfunc(*args, *results)

            for expected_i, result_i in zip(expected, results):
                msg = "\n".join(
                    [
                        "ufunc '{0}' failed",
                        "inputs ({1}):",
                        "{2}",
                        "got({3})",
                        "{4}",
                        "expected ({5}):",
                        "{6}",
                    ]
                ).format(
                    ufunc.__name__,
                    input_type,
                    input_operand,
                    output_type,
                    result_i,
                    expected_i.dtype,
                    expected_i,
                )
                try:
                    np.testing.assert_array_almost_equal(
                        expected_i, result_i, decimal=5, err_msg=msg
                    )
                except AssertionError:
                    if invalid_flag:
                        # Allow output to mismatch for invalid input
                        print(
                            "Output mismatch for invalid input",
                            input_tuple,
                            result_i,
                            expected_i,
                        )
                    else:
                        raise

    def signed_unsigned_cmp_test(self, comparison_ufunc):
        self.basic_ufunc_test(comparison_ufunc)

        if numpy_support.numpy_version < (1, 25):
            return

        # Test additional implementations that specifically handle signed /
        # unsigned comparisons added in NumPy 1.25:
        # https://github.com/numpy/numpy/pull/23713
        additional_inputs = (
            (np.int64(-1), np.uint64(0)),
            (np.int64(-1), np.uint64(1)),
            (np.int64(0), np.uint64(0)),
            (np.int64(0), np.uint64(1)),
            (np.int64(1), np.uint64(0)),
            (np.int64(1), np.uint64(1)),
            (np.uint64(0), np.int64(-1)),
            (np.uint64(0), np.int64(0)),
            (np.uint64(0), np.int64(1)),
            (np.uint64(1), np.int64(-1)),
            (np.uint64(1), np.int64(0)),
            (np.uint64(1), np.int64(1)),
            (
                np.array([-1, -1, 0, 0, 1, 1], dtype=np.int64),
                np.array([0, 1, 0, 1, 0, 1], dtype=np.uint64),
            ),
            (
                np.array([0, 1, 0, 1, 0, 1], dtype=np.uint64),
                np.array([-1, -1, 0, 0, 1, 1], dtype=np.int64),
            ),
        )

        pyfunc = self._make_ufunc_usecase(comparison_ufunc)

        for a, b in additional_inputs:
            input_types = (typeof(a), typeof(b))
            output_type = types.Array(types.bool_, 1, "C")
            argtys = input_types + (output_type,)
            cfunc = self._compile(pyfunc, argtys)

            if isinstance(a, np.ndarray):
                result = np.zeros(a.shape, dtype=np.bool_)
            else:
                result = np.zeros(1, dtype=np.bool_)

            expected = np.zeros_like(result)

            pyfunc(a, b, expected)
            cfunc(a, b, result)
            np.testing.assert_equal(expected, result)


def _make_ufunc_usecase(ufunc):
    ldict = {}
    arg_str = ",".join(["a{0}".format(i) for i in range(ufunc.nargs)])
    func_str = f"def fn({arg_str}):\n    np.{ufunc.__name__}({arg_str})"
    exec(func_str, globals(), ldict)
    fn = ldict["fn"]
    fn.__name__ = "{0}_usecase".format(ufunc.__name__)
    return fn


# This class provides common functionality for UFunc tests. The UFunc tests
# are quite long-running in comparison to other tests, so we break the tests up
# into multiple test classes for distribution across workers.
#
# This class would also be a CUDATestCase, but to avoid a confusing and
# potentially dangerous inheritance diamond with setUp methods that modify
# global state, we implement the necessary part of CUDATestCase within this
# class instead. This disables CUDA performance warnings for the duration of
# tests.
class CUDAUFuncTestBase(BasicUFuncTest, TestCase):
    def setUp(self):
        BasicUFuncTest.setUp(self)

        # The basic ufunc test does not set up complex inputs, so we'll add
        # some here for testing with CUDA.
        self.inputs.extend(
            [
                (np.complex64(-0.5 - 0.5j), types.complex64),
                (np.complex64(0.0), types.complex64),
                (np.complex64(0.5 + 0.5j), types.complex64),
                (np.complex128(-0.5 - 0.5j), types.complex128),
                (np.complex128(0.0), types.complex128),
                (np.complex128(0.5 + 0.5j), types.complex128),
                (
                    np.array([-0.5 - 0.5j, 0.0, 0.5 + 0.5j], dtype="c8"),
                    types.Array(types.complex64, 1, "C"),
                ),
                (
                    np.array([-0.5 - 0.5j, 0.0, 0.5 + 0.5j], dtype="c16"),
                    types.Array(types.complex128, 1, "C"),
                ),
            ]
        )

        # Test with multiple dimensions
        self.inputs.extend(
            [
                # Basic 2D and 3D arrays
                (
                    np.linspace(0, 1).reshape((5, -1)),
                    types.Array(types.float64, 2, "C"),
                ),
                (
                    np.linspace(0, 1).reshape((2, 5, -1)),
                    types.Array(types.float64, 3, "C"),
                ),
                # Complex data (i.e. interleaved)
                (
                    np.linspace(0, 1 + 1j).reshape(5, -1),
                    types.Array(types.complex128, 2, "C"),
                ),
                # F-ordered
                (
                    np.asfortranarray(np.linspace(0, 1).reshape((5, -1))),
                    types.Array(types.float64, 2, "F"),
                ),
            ]
        )

        # Add tests for other integer types
        self.inputs.extend(
            [
                (np.uint8(0), types.uint8),
                (np.uint8(1), types.uint8),
                (np.int8(-1), types.int8),
                (np.int8(0), types.int8),
                (np.uint16(0), types.uint16),
                (np.uint16(1), types.uint16),
                (np.int16(-1), types.int16),
                (np.int16(0), types.int16),
                (np.ulonglong(0), types.ulonglong),
                (np.ulonglong(1), types.ulonglong),
                (np.longlong(-1), types.longlong),
                (np.longlong(0), types.longlong),
                (
                    np.array([0, 1], dtype=np.ulonglong),
                    types.Array(types.ulonglong, 1, "C"),
                ),
                (
                    np.array([0, 1], dtype=np.longlong),
                    types.Array(types.longlong, 1, "C"),
                ),
            ]
        )

        self._low_occupancy_warnings = config.CUDA_LOW_OCCUPANCY_WARNINGS
        self._warn_on_implicit_copy = config.CUDA_WARN_ON_IMPLICIT_COPY

        # Disable warnings about low gpu utilization in the test suite
        config.CUDA_LOW_OCCUPANCY_WARNINGS = 0
        # Disable warnings about host arrays in the test suite
        config.CUDA_WARN_ON_IMPLICIT_COPY = 0

    def tearDown(self):
        # Restore original warning settings
        config.CUDA_LOW_OCCUPANCY_WARNINGS = self._low_occupancy_warnings
        config.CUDA_WARN_ON_IMPLICIT_COPY = self._warn_on_implicit_copy

    def _make_ufunc_usecase(self, ufunc):
        return _make_ufunc_usecase(ufunc)

    @functools.lru_cache(maxsize=None)
    def _compile(self, pyfunc, args):
        # We return an already-configured kernel so that basic_ufunc_test can
        # call it just like it does for a CPU function
        return cuda.jit(args)(pyfunc)[1, 1]

    def basic_int_ufunc_test(self, name=None):
        skip_inputs = [
            types.float32,
            types.float64,
            types.Array(types.float32, 1, "C"),
            types.Array(types.float32, 2, "C"),
            types.Array(types.float64, 1, "C"),
            types.Array(types.float64, 2, "C"),
            types.Array(types.float64, 3, "C"),
            types.Array(types.float64, 2, "F"),
            types.complex64,
            types.complex128,
            types.Array(types.complex64, 1, "C"),
            types.Array(types.complex64, 2, "C"),
            types.Array(types.complex128, 1, "C"),
            types.Array(types.complex128, 2, "C"),
        ]
        self.basic_ufunc_test(name, skip_inputs=skip_inputs)

    ############################################################################
    # Trigonometric Functions


class TestBasicTrigUFuncs(CUDAUFuncTestBase):
    def test_sin_ufunc(self):
        self.basic_ufunc_test(np.sin, kinds="cf")

    def test_cos_ufunc(self):
        self.basic_ufunc_test(np.cos, kinds="cf")

    def test_tan_ufunc(self):
        self.basic_ufunc_test(np.tan, kinds="cf")

    def test_arcsin_ufunc(self):
        self.basic_ufunc_test(np.arcsin, kinds="cf")

    def test_arccos_ufunc(self):
        self.basic_ufunc_test(np.arccos, kinds="cf")

    def test_arctan_ufunc(self):
        self.basic_ufunc_test(np.arctan, kinds="cf")

    def test_arctan2_ufunc(self):
        self.basic_ufunc_test(np.arctan2, kinds="f")


class TestHypTrigUFuncs(CUDAUFuncTestBase):
    def test_hypot_ufunc(self):
        self.basic_ufunc_test(np.hypot, kinds="f")

    def test_sinh_ufunc(self):
        self.basic_ufunc_test(np.sinh, kinds="cf")

    def test_cosh_ufunc(self):
        self.basic_ufunc_test(np.cosh, kinds="cf")

    def test_tanh_ufunc(self):
        self.basic_ufunc_test(np.tanh, kinds="cf")

    def test_arcsinh_ufunc(self):
        self.basic_ufunc_test(np.arcsinh, kinds="cf")

    def test_arccosh_ufunc(self):
        self.basic_ufunc_test(np.arccosh, kinds="cf")

    def test_arctanh_ufunc(self):
        # arctanh is only valid is only finite in the range ]-1, 1[
        # This means that for any of the integer types it will produce
        # conversion from infinity/-infinity to integer. That's undefined
        # behavior in C, so the results may vary from implementation to
        # implementation. This means that the result from the compiler
        # used to compile NumPy may differ from the result generated by
        # llvm. Skipping the integer types in this test avoids failed
        # tests because of this.
        to_skip = [
            types.Array(types.uint32, 1, "C"),
            types.uint32,
            types.Array(types.int32, 1, "C"),
            types.int32,
            types.Array(types.uint64, 1, "C"),
            types.uint64,
            types.Array(types.int64, 1, "C"),
            types.int64,
        ]

        self.basic_ufunc_test(np.arctanh, skip_inputs=to_skip, kinds="cf")


class TestConversionUFuncs(CUDAUFuncTestBase):
    def test_deg2rad_ufunc(self):
        self.basic_ufunc_test(np.deg2rad, kinds="f")

    def test_rad2deg_ufunc(self):
        self.basic_ufunc_test(np.rad2deg, kinds="f")

    def test_degrees_ufunc(self):
        self.basic_ufunc_test(np.degrees, kinds="f")

    def test_radians_ufunc(self):
        self.basic_ufunc_test(np.radians, kinds="f")

    ############################################################################
    # Comparison functions


class TestComparisonUFuncs1(CUDAUFuncTestBase):
    def test_greater_ufunc(self):
        self.signed_unsigned_cmp_test(np.greater)

    def test_greater_equal_ufunc(self):
        self.signed_unsigned_cmp_test(np.greater_equal)

    def test_less_ufunc(self):
        self.signed_unsigned_cmp_test(np.less)

    def test_less_equal_ufunc(self):
        self.signed_unsigned_cmp_test(np.less_equal)

    def test_not_equal_ufunc(self):
        self.signed_unsigned_cmp_test(np.not_equal)

    def test_equal_ufunc(self):
        self.signed_unsigned_cmp_test(np.equal)


class TestLogicalUFuncs(CUDAUFuncTestBase):
    def test_logical_and_ufunc(self):
        self.basic_ufunc_test(np.logical_and)

    def test_logical_or_ufunc(self):
        self.basic_ufunc_test(np.logical_or)

    def test_logical_xor_ufunc(self):
        self.basic_ufunc_test(np.logical_xor)

    def test_logical_not_ufunc(self):
        self.basic_ufunc_test(np.logical_not)


class TestMinmaxUFuncs(CUDAUFuncTestBase):
    def test_maximum_ufunc(self):
        self.basic_ufunc_test(np.maximum)

    def test_minimum_ufunc(self):
        self.basic_ufunc_test(np.minimum)

    def test_fmax_ufunc(self):
        self.basic_ufunc_test(np.fmax)

    def test_fmin_ufunc(self):
        self.basic_ufunc_test(np.fmin)


class TestBitwiseUFuncs(CUDAUFuncTestBase):
    def test_bitwise_and_ufunc(self):
        self.basic_int_ufunc_test(np.bitwise_and)

    def test_bitwise_or_ufunc(self):
        self.basic_int_ufunc_test(np.bitwise_or)

    def test_bitwise_xor_ufunc(self):
        self.basic_int_ufunc_test(np.bitwise_xor)

    def test_invert_ufunc(self):
        self.basic_int_ufunc_test(np.invert)

    def test_bitwise_not_ufunc(self):
        self.basic_int_ufunc_test(np.bitwise_not)

    # Note: there is no entry for np.left_shift and np.right_shift
    # because their implementations in NumPy have undefined behavior
    # when the second argument is a negative. See the comment in
    # numba/tests/test_ufuncs.py for more details.

    ############################################################################
    # Mathematical Functions


class TestLogUFuncs(CUDAUFuncTestBase):
    def test_log_ufunc(self):
        self.basic_ufunc_test(np.log, kinds="cf")

    def test_log2_ufunc(self):
        self.basic_ufunc_test(np.log2, kinds="cf")

    def test_log10_ufunc(self):
        self.basic_ufunc_test(np.log10, kinds="cf")


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import concurrent.futures
import multiprocessing
import os
from numba.cuda.testing import unittest


def set_visible_devices_and_check():
    from numba import cuda
    import os

    os.environ["CUDA_VISIBLE_DEVICES"] = "0"
    return len(cuda.gpus.lst)


class TestVisibleDevices(unittest.TestCase):
    def test_visible_devices_set_after_import(self):
        # See Issue #6149. This test checks that we can set
        # CUDA_VISIBLE_DEVICES after importing Numba and have the value
        # reflected in the available list of GPUs. Prior to the fix for this
        # issue, Numba made a call to runtime.get_version() on import that
        # initialized the driver and froze the list of available devices before
        # CUDA_VISIBLE_DEVICES could be set by the user.

        # Avoid importing cuda at the top level so that
        # set_visible_devices_and_check gets to import it first in its process
        from numba import cuda

        if len(cuda.gpus.lst) in (0, 1):
            self.skipTest("This test requires multiple GPUs")

        if os.environ.get("CUDA_VISIBLE_DEVICES"):
            msg = "Cannot test when CUDA_VISIBLE_DEVICES already set"
            self.skipTest(msg)

        with concurrent.futures.ProcessPoolExecutor(
            mp_context=multiprocessing.get_context("spawn")
        ) as exe:
            future = exe.submit(set_visible_devices_and_check)

        visible_gpu_count = future.result()
        assert visible_gpu_count == 1


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import os
import multiprocessing as mp
import pytest
import concurrent.futures

import numpy as np

from numba import cuda
from numba.cuda.testing import skip_on_cudasim, CUDATestCase
import unittest


@skip_on_cudasim("disabled for cudasim")
class TestMultiprocessing(CUDATestCase):
    @unittest.skipUnless(hasattr(mp, "get_context"), "requires mp.get_context")
    @unittest.skipUnless(os.name == "posix", "requires Unix")
    def test_fork(self):
        """
        Test fork detection.
        """
        from numba.cuda.cudadrv.error import CudaDriverError

        cuda.current_context()  # force cuda initialize
        with concurrent.futures.ProcessPoolExecutor(
            mp_context=mp.get_context("fork")
        ) as exe:
            future = exe.submit(cuda.to_device, np.arange(1))

        with pytest.raises(
            CudaDriverError, match="CUDA initialized before forking"
        ):
            future.result()


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

# Tests numba.analysis functions
import collections
import types as pytypes

import numpy as np
from numba.cuda.compiler import run_frontend
from numba.cuda.flags import Flags
from numba.cuda.core.compiler import StateDict
from numba.cuda import jit
from numba.cuda import types
from numba.cuda.core import errors
from numba.cuda.core import ir
from numba.cuda.utils import PYVERSION
from numba.cuda.core import postproc, rewrites, ir_utils
from numba.cuda.core.options import ParallelOptions
from numba.cuda.core.inline_closurecall import InlineClosureCallPass
from numba.cuda.tests.support import TestCase, override_config
from numba.cuda.core.analysis import (
    dead_branch_prune,
    rewrite_semantic_constants,
)
from numba.cuda.core.untyped_passes import (
    ReconstructSSA,
)
import unittest
from numba.cuda.core import config

_GLOBAL = 123

enable_pyobj_flags = Flags()
enable_pyobj_flags.enable_pyobject = True

if config.ENABLE_CUDASIM:
    raise unittest.SkipTest("Analysis passes not done in simulator")


def compile_to_ir(func):
    func_ir = run_frontend(func)
    state = StateDict()
    state.func_ir = func_ir
    state.typemap = None
    state.calltypes = None
    # Transform to SSA
    ReconstructSSA().run_pass(state)
    # call this to get print etc rewrites
    rewrites.rewrite_registry.apply("before-inference", state)
    return func_ir


class TestBranchPruneBase(TestCase):
    """
    Tests branch pruning
    """

    _DEBUG = False

    # find *all* branches
    def find_branches(self, the_ir):
        branches = []
        for blk in the_ir.blocks.values():
            tmp = [_ for _ in blk.find_insts(cls=ir.Branch)]
            branches.extend(tmp)
        return branches

    def assert_prune(self, func, args_tys, prune, *args, **kwargs):
        # This checks that the expected pruned branches have indeed been pruned.
        # func is a python function to assess
        # args_tys is the numba types arguments tuple
        # prune arg is a list, one entry per branch. The value in the entry is
        # encoded as follows:
        # True: using constant inference only, the True branch will be pruned
        # False: using constant inference only, the False branch will be pruned
        # None: under no circumstances should this branch be pruned
        # *args: the argument instances to pass to the function to check
        #        execution is still valid post transform
        # **kwargs:
        #        - flags: args to pass to `jit` default is `nopython=True`,
        #          e.g. permits use of e.g. object mode.

        func_ir = compile_to_ir(func)
        before = func_ir.copy()
        if self._DEBUG:
            print("=" * 80)
            print("before inline")
            func_ir.dump()

        # run closure inlining to ensure that nonlocals in closures are visible
        inline_pass = InlineClosureCallPass(
            func_ir,
            ParallelOptions(False),
        )
        inline_pass.run()

        # Remove all Dels, and re-run postproc
        post_proc = postproc.PostProcessor(func_ir)
        post_proc.run()

        rewrite_semantic_constants(func_ir, args_tys)
        if self._DEBUG:
            print("=" * 80)
            print("before prune")
            func_ir.dump()

        dead_branch_prune(func_ir, args_tys)

        after = func_ir
        if self._DEBUG:
            print("after prune")
            func_ir.dump()

        before_branches = self.find_branches(before)
        self.assertEqual(len(before_branches), len(prune))

        # what is expected to be pruned
        expect_removed = []
        for idx, prune in enumerate(prune):
            branch = before_branches[idx]
            if prune is True:
                expect_removed.append(branch.truebr)
            elif prune is False:
                expect_removed.append(branch.falsebr)
            elif prune is None:
                pass  # nothing should be removed!
            elif prune == "both":
                expect_removed.append(branch.falsebr)
                expect_removed.append(branch.truebr)
            else:
                assert 0, "unreachable"

        # compare labels
        original_labels = set([_ for _ in before.blocks.keys()])
        new_labels = set([_ for _ in after.blocks.keys()])
        # assert that the new labels are precisely the original less the
        # expected pruned labels
        try:
            self.assertEqual(new_labels, original_labels - set(expect_removed))
        except AssertionError as e:
            print("new_labels", sorted(new_labels))
            print("original_labels", sorted(original_labels))
            print("expect_removed", sorted(expect_removed))
            raise e

        if [
            arg is types.NoneType("none") or arg is types.Omitted(None)
            for arg in args_tys
        ].count(True) == 0:
            self.run_func(func, args)

    def run_func(self, impl, args):
        cres = jit(impl)
        dargs = args
        out = np.zeros(1)
        cout = np.zeros(1)
        args += (out,)
        dargs += (cout,)
        cres.py_func(*args)
        with override_config("DISABLE_PERFORMANCE_WARNINGS", 1):
            cres[1, 1](*dargs)
        self.assertPreciseEqual(out[0], cout[0])


class TestBranchPrune(TestBranchPruneBase):
    def test_single_if(self):
        def impl(x, res):
            if 1 == 0:
                res[0] = 3.14159

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [True],
            None,
        )

        def impl(x, res):
            if 1 == 1:
                res[0] = 3.14159

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [False],
            None,
        )

        def impl(x, res):
            if x is None:
                res[0] = 3.14159

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [False],
            None,
        )
        self.assert_prune(
            impl,
            (types.IntegerLiteral(10), types.Array(types.float64, 1, "C")),
            [True],
            10,
        )

        def impl(x, res):
            if x == 10:
                res[0] = 3.14159

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [True],
            None,
        )
        self.assert_prune(
            impl,
            (types.IntegerLiteral(10), types.Array(types.float64, 1, "C")),
            [None],
            10,
        )

        def impl(x, res):
            if x == 10:
                z = 3.14159  # noqa: F841 # no effect

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [True],
            None,
        )
        self.assert_prune(
            impl,
            (types.IntegerLiteral(10), types.Array(types.float64, 1, "C")),
            [None],
            10,
        )

        def impl(x, res):
            z = None
            y = z
            if x == y:
                res[0] = 100

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [False],
            None,
        )
        self.assert_prune(
            impl,
            (types.IntegerLiteral(10), types.Array(types.float64, 1, "C")),
            [True],
            10,
        )

    def test_single_if_else(self):
        def impl(x, res):
            if x is None:
                res[0] = 3.14159
            else:
                res[0] = 1.61803

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [False],
            None,
        )
        self.assert_prune(
            impl,
            (types.IntegerLiteral(10), types.Array(types.float64, 1, "C")),
            [True],
            10,
        )

    def test_single_if_const_val(self):
        def impl(x, res):
            if x == 100:
                res[0] = 3.14159

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [True],
            None,
        )
        self.assert_prune(
            impl,
            (types.IntegerLiteral(100), types.Array(types.float64, 1, "C")),
            [None],
            100,
        )

        def impl(x, res):
            # switch the condition order
            if 100 == x:
                res[0] = 3.14159

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [True],
            None,
        )
        self.assert_prune(
            impl,
            (types.IntegerLiteral(100), types.Array(types.float64, 1, "C")),
            [None],
            100,
        )

    def test_single_if_else_two_const_val(self):
        def impl(x, y, res):
            if x == y:
                res[0] = 3.14159
            else:
                res[0] = 1.61803

        self.assert_prune(
            impl,
            (types.IntegerLiteral(100),) * 2
            + (types.Array(types.float64, 1, "C"),),
            [None],
            100,
            100,
        )
        self.assert_prune(
            impl,
            (types.NoneType("none"),) * 2
            + (types.Array(types.float64, 1, "C"),),
            [False],
            None,
            None,
        )
        self.assert_prune(
            impl,
            (
                types.IntegerLiteral(100),
                types.NoneType("none"),
                types.Array(types.float64, 1, "C"),
            ),
            [True],
            100,
            None,
        )
        self.assert_prune(
            impl,
            (
                types.IntegerLiteral(100),
                types.IntegerLiteral(1000),
                types.Array(types.float64, 1, "C"),
            ),
            [None],
            100,
            1000,
        )

    def test_single_if_else_w_following_undetermined(self):
        def impl(x, res):
            x_is_none_work = False
            if x is None:
                x_is_none_work = True
            else:
                dead = 7  # noqa: F841 # no effect

            if x_is_none_work:
                y = 10
            else:
                y = -3
            res[0] = y

        self.assert_prune(
            impl,
            (types.NoneType("none"), types.Array(types.float64, 1, "C")),
            [False, None],
            None,
        )
        self.assert_prune(
            impl,
            (types.IntegerLiteral(10), types.Array(types.float64, 1, "C")),
            [True, None],
            10,
        )

        def impl(x, res):
            x_is_none_work = False
            if x is None:
                x_is_none_work = True
            else:
                pass

            if x_is_none_work:
                y = 10
            else:
                y = -3
            res[0] = y

        # Python 3.10 creates a block with a NOP in it for the `pass` which
        # means it gets pruned.
        if PYVERSION >= (3, 10):
            # Python 3.10 creates a block with a NOP in it for the `pass` which
            # means it gets pruned.
            self.assert_prune(
                impl,
                (types.NoneType("none"), types.Array(types.float64, 1, "C")),
                [False, None],
                None,
            )
        else:
            self.assert_prune(
                impl,
                (types.NoneType("none"), types.Array(types.float64, 1, "C")),
                [None, None],
                None,
            )

        self.assert_prune(
            impl,
            (types.IntegerLiteral(10), types.Array(types.float64, 1, "C")),
            [True, None],
            10,
        )

    def test_double_if_else_rt_const(self):
        def impl(x, res):
            one_hundred = 100
            x_is_none_work = 4
            if x is None:
                x_is_none_work = 100
            else:
                dead = 7  # noqa: F841 # no effect

            if x_is_none_work == one_hundred:
                y = 10
            else:
                y = -3

            res[0] = y + x_is_none_work

        self.assert_prune(impl, (types.NoneType("none"),), [False, None], None)
        self.assert_prune(impl, (types.IntegerLiteral(10),), [True, None], 10)

    def test_double_if_else_non_literal_const(self):
        def impl(x, res):
            one_hundred = 100
            if x == one_hundred:
                res[0] = 3.14159
            else:
                res[0] = 1.61803

        # no prune as compilation specialization on literal value not permitted
        self.assert_prune(impl, (types.IntegerLiteral(10),), [None], 10)
        self.assert_prune(impl, (types.IntegerLiteral(100),), [None], 100)

    def test_single_two_branches_same_cond(self):
        def impl(x, res):
            if x is None:
                y = 10
            else:
                y = 40

            if x is not None:
                z = 100
            else:
                z = 400

            res[0] = z + y

        self.assert_prune(impl, (types.NoneType("none"),), [False, True], None)
        self.assert_prune(impl, (types.IntegerLiteral(10),), [True, False], 10)

    def test_cond_is_kwarg_none(self):
        def impl(x=None, res=None):
            if x is None:
                y = 10
            else:
                y = 40

            if x is not None:
                z = 100
            else:
                z = 400

            res[0] = z + y

        self.assert_prune(impl, (types.Omitted(None),), [False, True], None)
        self.assert_prune(impl, (types.NoneType("none"),), [False, True], None)
        self.assert_prune(impl, (types.IntegerLiteral(10),), [True, False], 10)

    def test_cond_is_kwarg_value(self):
        def impl(x=1000, res=None):
            if x == 1000:
                y = 10
            else:
                y = 40

            if x != 1000:
                z = 100
            else:
                z = 400

            res[0] = z + y

        self.assert_prune(impl, (types.Omitted(1000),), [None, None], 1000)
        self.assert_prune(
            impl, (types.IntegerLiteral(1000),), [None, None], 1000
        )
        self.assert_prune(impl, (types.IntegerLiteral(0),), [None, None], 0)
        self.assert_prune(impl, (types.NoneType("none"),), [True, False], None)

    def test_cond_rewrite_is_correct(self):
        # this checks that when a condition is replaced, it is replace by a
        # true/false bit that correctly represents the evaluated condition
        def fn(x):
            if x is None:
                return 10
            return 12

        def check(func, arg_tys, bit_val):
            func_ir = compile_to_ir(func)

            # check there is 1 branch
            before_branches = self.find_branches(func_ir)
            self.assertEqual(len(before_branches), 1)

            # check the condition in the branch is a binop
            pred_var = before_branches[0].cond
            pred_defn = ir_utils.get_definition(func_ir, pred_var)
            self.assertEqual(pred_defn.op, "call")
            condition_var = pred_defn.args[0]
            condition_op = ir_utils.get_definition(func_ir, condition_var)
            self.assertEqual(condition_op.op, "binop")

            # do the prune, this should kill the dead branch and rewrite the
            #'condition to a true/false const bit
            if self._DEBUG:
                print("=" * 80)
                print("before prune")
                func_ir.dump()
            dead_branch_prune(func_ir, arg_tys)
            if self._DEBUG:
                print("=" * 80)
                print("after prune")
                func_ir.dump()

            # after mutation, the condition should be a const value `bit_val`
            new_condition_defn = ir_utils.get_definition(func_ir, condition_var)
            self.assertTrue(isinstance(new_condition_defn, ir.Const))
            self.assertEqual(new_condition_defn.value, bit_val)

        check(fn, (types.NoneType("none"),), 1)
        check(fn, (types.IntegerLiteral(10),), 0)

    def test_global_bake_in(self):
        def impl(x, res):
            if _GLOBAL == 123:
                res[0] = x
            else:
                res[0] = x + 10

        self.assert_prune(
            impl,
            (types.IntegerLiteral(1), types.Array(types.float64, 1, "C")),
            [False],
            1,
        )

        global _GLOBAL
        tmp = _GLOBAL

        try:
            _GLOBAL = 5

            def impl(x, res):
                if _GLOBAL == 123:
                    res[0] = x
                else:
                    res[0] = x + 10

            self.assert_prune(
                impl,
                (types.IntegerLiteral(1), types.Array(types.float64, 1, "C")),
                [True],
                1,
            )
        finally:
            _GLOBAL = tmp

    def test_freevar_bake_in(self):
        _FREEVAR = 123

        def impl(x, res):
            if _FREEVAR == 123:
                res[0] = x
            else:
                res[0] = x + 10

        self.assert_prune(
            impl,
            (types.IntegerLiteral(1), types.Array(types.float64, 1, "C")),
            [False],
            1,
        )

        _FREEVAR = 12

        def impl(x, res):
            if _FREEVAR == 123:
                res[0] = x
            else:
                res[0] = x + 10

        self.assert_prune(
            impl,
            (types.IntegerLiteral(1), types.Array(types.float64, 1, "C")),
            [True],
            1,
        )

    def test_redefined_variables_are_not_considered_in_prune(self):
        # see issue #4163, checks that if a variable that is an argument is
        # redefined in the user code it is not considered const

        def impl(array, a=None, res=None):
            if a is None:
                a = 0
            if a < 0:
                res[0] = 10
            res[0] = 30

        self.assert_prune(
            impl,
            (
                types.Array(types.float64, 2, "C"),
                types.NoneType("none"),
                types.Array(types.float64, 1, "C"),
            ),
            [None, None],
            np.zeros((2, 3)),
            None,
        )

    def test_redefinition_analysis_same_block(self):
        # checks that a redefinition in a block with prunable potential doesn't
        # break

        def impl(array, x, a=None, res=None):
            b = 2
            if x < 4:
                b = 12
            if a is None:  # known true
                a = 7  # live
            else:
                b = 15  # dead
            if a < 0:  # valid as a result of the redefinition of 'a'
                res[0] = 10
            res[0] = 30 + b + a

        self.assert_prune(
            impl,
            (
                types.Array(types.float64, 2, "C"),
                types.float64,
                types.NoneType("none"),
                types.Array(types.float64, 1, "C"),
            ),
            [None, False, None],
            np.zeros((2, 3)),
            1.0,
            None,
        )

    def test_redefinition_analysis_different_block_can_exec(self):
        # checks that a redefinition in a block that may be executed prevents
        # pruning

        def impl(array, x, res):
            b = 0
            if x > 5:
                a = 11  # a redefined, cannot tell statically if this will exec
            if x < 4:
                b = 12
            if a is None:  # cannot prune, cannot determine if re-defn occurred
                b += 5
            else:
                b += 7
                if a < 0:
                    res[0] = 10
            res[0] = 30 + b

        self.assert_prune(
            impl,
            (
                types.Array(types.float64, 2, "C"),
                types.float64,
                types.NoneType("none"),
                types.Array(types.float64, 1, "C"),
            ),
            [None, None, None, None],
            np.zeros((2, 3)),
            1.0,
            None,
        )

    def test_redefinition_analysis_different_block_cannot_exec(self):
        # checks that a redefinition in a block guarded by something that
        # has prune potential

        def impl(array, x=None, a=None, res=None):
            b = 0
            if x is not None:
                a = 11
            if a is None:
                b += 5
            else:
                b += 7
            res[0] = 30 + b

        self.assert_prune(
            impl,
            (
                types.Array(types.float64, 2, "C"),
                types.NoneType("none"),
                types.NoneType("none"),
                types.Array(types.float64, 1, "C"),
            ),
            [True, None],
            np.zeros((2, 3)),
            None,
            None,
        )

        self.assert_prune(
            impl,
            (
                types.Array(types.float64, 2, "C"),
                types.NoneType("none"),
                types.float64,
                types.Array(types.float64, 1, "C"),
            ),
            [True, None],
            np.zeros((2, 3)),
            None,
            1.2,
        )

        self.assert_prune(
            impl,
            (
                types.Array(types.float64, 2, "C"),
                types.float64,
                types.NoneType("none"),
                types.Array(types.float64, 1, "C"),
            ),
            [None, None],
            np.zeros((2, 3)),
            1.2,
            None,
        )

    def test_closure_and_nonlocal_can_prune(self):
        # Closures must be inlined ahead of branch pruning in case nonlocal
        # is used. See issue #6585.
        def impl(res):
            x = 1000

            def closure():
                nonlocal x
                x = 0

            closure()

            if x == 0:
                res[0] = True
            else:
                res[0] = False

        self.assert_prune(
            impl,
            (types.Array(types.float64, 1, "C"),),
            [
                False,
            ],
        )

    def test_closure_and_nonlocal_cannot_prune(self):
        # Closures must be inlined ahead of branch pruning in case nonlocal
        # is used. See issue #6585.
        def impl(n, res):
            x = 1000

            def closure(t):
                nonlocal x
                x = t

            closure(n)

            if x == 0:
                res[0] = True
            else:
                res[0] = False

        self.assert_prune(
            impl,
            (types.int64, types.Array(types.float64, 1, "C")),
            [
                None,
            ],
            1,
        )


class TestBranchPrunePredicates(TestBranchPruneBase):
    # Really important thing to remember... the branch on predicates end up as
    # POP_JUMP_IF_<bool> and the targets are backwards compared to normal, i.e.
    # the true condition is far jump and the false the near i.e. `if x` would
    # end up in Numba IR as e.g. `branch x 10, 6`.

    _TRUTHY = (1, "String", True, 7.4, 3j)
    _FALSEY = (0, "", False, 0.0, 0j, None)

    def _literal_const_sample_generator(self, pyfunc, consts):
        """
        This takes a python function, pyfunc, and manipulates its co_const
        __code__ member to create a new function with different co_consts as
        supplied in argument consts.

        consts is a dict {index: value} of co_const tuple index to constant
        value used to update a pyfunc clone's co_const.
        """
        pyfunc_code = pyfunc.__code__

        # translate consts spec to update the constants
        co_consts = {k: v for k, v in enumerate(pyfunc_code.co_consts)}
        for k, v in consts.items():
            co_consts[k] = v
        new_consts = tuple([v for _, v in sorted(co_consts.items())])

        # create code object with mutation
        new_code = pyfunc_code.replace(co_consts=new_consts)

        # get function
        return pytypes.FunctionType(new_code, globals())

    def test_literal_const_code_gen(self):
        def impl(x):
            _CONST1 = "PLACEHOLDER1"
            if _CONST1:
                return 3.14159
            else:
                _CONST2 = "PLACEHOLDER2"
            return _CONST2 + 4

        new = self._literal_const_sample_generator(impl, {1: 0, 3: 20})
        iconst = impl.__code__.co_consts
        nconst = new.__code__.co_consts
        self.assertEqual(
            iconst, (None, "PLACEHOLDER1", 3.14159, "PLACEHOLDER2", 4)
        )
        self.assertEqual(nconst, (None, 0, 3.14159, 20, 4))
        self.assertEqual(impl(None), 3.14159)
        self.assertEqual(new(None), 24)

    def test_single_if_const(self):
        def impl(x):
            _CONST1 = "PLACEHOLDER1"
            if _CONST1:
                return 3.14159

        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for const in c_inp:
                func = self._literal_const_sample_generator(impl, {1: const})
                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_negate_const(self):
        def impl(x):
            _CONST1 = "PLACEHOLDER1"
            if not _CONST1:
                return 3.14159

        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for const in c_inp:
                func = self._literal_const_sample_generator(impl, {1: const})
                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_else_const(self):
        def impl(x):
            _CONST1 = "PLACEHOLDER1"
            if _CONST1:
                return 3.14159
            else:
                return 1.61803

        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for const in c_inp:
                func = self._literal_const_sample_generator(impl, {1: const})
                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_else_negate_const(self):
        def impl(x):
            _CONST1 = "PLACEHOLDER1"
            if not _CONST1:
                return 3.14159
            else:
                return 1.61803

        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for const in c_inp:
                func = self._literal_const_sample_generator(impl, {1: const})
                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_freevar(self):
        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for const in c_inp:

                def func(x):
                    if const:
                        return 3.14159, const

                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_negate_freevar(self):
        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for const in c_inp:

                def func(x):
                    if not const:
                        return 3.14159, const

                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_else_negate_freevar(self):
        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for const in c_inp:

                def func(x):
                    if not const:
                        return 3.14159, const
                    else:
                        return 1.61803, const

                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    # globals in this section have absurd names after their test usecase names
    # so as to prevent collisions and permit tests to run in parallel
    def test_single_if_global(self):
        global c_test_single_if_global

        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for c in c_inp:
                c_test_single_if_global = c

                def func(x):
                    if c_test_single_if_global:
                        return 3.14159, c_test_single_if_global

                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_negate_global(self):
        global c_test_single_if_negate_global

        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for c in c_inp:
                c_test_single_if_negate_global = c

                def func(x):
                    if c_test_single_if_negate_global:
                        return 3.14159, c_test_single_if_negate_global

                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_else_global(self):
        global c_test_single_if_else_global

        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for c in c_inp:
                c_test_single_if_else_global = c

                def func(x):
                    if c_test_single_if_else_global:
                        return 3.14159, c_test_single_if_else_global
                    else:
                        return 1.61803, c_test_single_if_else_global

                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_single_if_else_negate_global(self):
        global c_test_single_if_else_negate_global

        for c_inp, prune in (self._TRUTHY, False), (self._FALSEY, True):
            for c in c_inp:
                c_test_single_if_else_negate_global = c

                def func(x):
                    if not c_test_single_if_else_negate_global:
                        return 3.14159, c_test_single_if_else_negate_global
                    else:
                        return 1.61803, c_test_single_if_else_negate_global

                self.assert_prune(
                    func, (types.NoneType("none"),), [prune], None
                )

    def test_issue_5618(self):
        @jit
        def foo(res):
            tmp = 666
            if tmp:
                res[0] = tmp

        self.run_func(foo, ())


class TestBranchPrunePostSemanticConstRewrites(TestBranchPruneBase):
    # Tests that semantic constants rewriting works by virtue of branch pruning

    def test_array_ndim_attr(self):
        def impl(array, res):
            if array.ndim == 2:
                if array.shape[1] == 2:
                    res[0] = 1
            else:
                res[0] = 10

        self.assert_prune(
            impl,
            (types.Array(types.float64, 2, "C"),),
            [False, None],
            np.zeros((2, 3)),
        )
        self.assert_prune(
            impl,
            (types.Array(types.float64, 1, "C"),),
            [True, "both"],
            np.zeros((2,)),
        )

    def test_tuple_len(self):
        def impl(tup, res):
            if len(tup) == 3:
                if tup[2] == 2:
                    res[0] = 1
            else:
                res[0] = 0

        self.assert_prune(
            impl,
            (types.UniTuple(types.int64, 3),),
            [False, None],
            tuple([1, 2, 3]),
        )
        self.assert_prune(
            impl,
            (types.UniTuple(types.int64, 2),),
            [True, "both"],
            tuple([1, 2]),
        )

    def test_attr_not_len(self):
        # The purpose of this test is to make sure that the conditions guarding
        # the rewrite part do not themselves raise exceptions.
        # This produces an `ir.Expr` call node for `float.as_integer_ratio`,
        # which is a getattr() on `float`.

        @jit
        def test():
            float.as_integer_ratio(1.23)

        # this should raise a TypingError
        with self.assertRaises(errors.TypingError) as e:
            test[1, 1]()

        self.assertIn("Unknown attribute 'as_integer_ratio'", str(e.exception))

    def test_ndim_not_on_array(self):
        FakeArray = collections.namedtuple("FakeArray", ["ndim"])
        fa = FakeArray(ndim=2)

        def impl(fa, res):
            if fa.ndim == 2:
                res[0] = fa.ndim

        # check prune works for array ndim
        self.assert_prune(
            impl,
            (types.Array(types.float64, 2, "C"),),
            [False],
            np.zeros((2, 3)),
        )

        # check prune fails for something with `ndim` attr that is not array
        FakeArrayType = types.NamedUniTuple(types.int64, 1, FakeArray)
        self.assert_prune(
            impl,
            (FakeArrayType,),
            [None],
            fa,
            flags={"nopython": False, "forceobj": True},
        )
